;;; -*- TI-Asm -*-
;;;
;;; DuckSign - sign Flash applications on the TI-83 Plus
;;;
;;; Copyright (C) 2010 Benjamin Moody
;;;
;;; This program is free software: you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;; AddN:
;;
;; Add two unsigned integers. The output buffer may be the same as one
;; or both input buffers, but must not partially overlap them.
;;
;; Input:
;; - A = size of numbers, bytes
;; - BC = address of first number (A bytes)
;; - HL = address of second number (A bytes)
;; - DE = address of destination (A bytes)
;;
;; Output:
;; - Carry flag set if MS addition carries
;;
;; Destroys:
;; - AF, BC', DE', HL'

AddModulus:
	ld	hl, (modulus)
	ld	a, (modulusLength)
AddN:
	push	hl
	 push	de
	  push	bc
	   exx
	   ld	b, a
	   or	a
AddN_Loop:
	   exx
	   ld	a, (bc)
	   adc	a, (hl)
	   ld	(de), a
	   inc	bc
	   inc	de
	   inc	hl
	   exx
	   djnz	AddN_Loop
	   pop	bc
	  pop	de
	 pop	hl
	ret

;; SubN:
;;
;; Subtract two unsigned integers.  The output buffer may be the same
;; as one or both input buffers, but must not partially overlap them.
;;
;; Input:
;; - A = size of numbers, bytes
;; - BC = address of first number (A bytes)
;; - HL = address of second number (A bytes)
;; - DE = address of destination (A bytes)
;;
;; Output:
;; - Carry flag set if MS subtraction carries
;;
;; Destroys:
;; - AF, BC', DE', HL'

SubModulus:
	ld	hl, (modulus)
SubML:
	ld	a, (modulusLength)
SubN:
	push	hl
	 push	de
	  push	bc
	   exx
	   ld	b, a
	   or	a
SubN_Loop:
	   exx
	   ld	a, (bc)
	   sbc	a, (hl)
	   ld	(de), a
	   inc	bc
	   inc	de
	   inc	hl
	   exx
	   djnz	SubN_Loop
	   pop	bc
	  pop	de
	 pop	hl
	ret

;; AddNtoM:
;;
;; Add two unsigned integers.
;;
;; Input:
;; - B = size of first number, bytes
;; - C = number of carry bytes (size of second number minus size of
;;   first; must be at least 1)
;; - HL = address of first number (B bytes, no carries)
;; - DE = address of second number and destination (B + C bytes, no
;;   carries)
;;
;; Output:
;; - DE increased by B bytes
;; - HL increased by B bytes
;;
;; Destroys:
;; - AF, BC

AddNtoM:
	or	a
AddNtoM_loop:
	ld	a, (de)
	adc	a, (hl)
	ld	(de), a
	inc	e
	FAIL_IF_Z
	inc	l
	FAIL_IF_Z
	djnz	AddNtoM_loop
	ret	nc
	push	hl
	 ld	h, d
	 ld	l, e
AddNtoM_carry:
	 inc	(hl)
	 jp	nz, AddNtoM_carry_done
	 inc	l
	 FAIL_IF_Z
	 dec	c
	 jr	nz, AddNtoM_carry
AddNtoM_carry_done:
	 pop	hl
	ret

