;;; -*- TI-Asm -*-
;;;
;;; DuckSign - sign Flash applications on the TI-83 Plus
;;;
;;; Copyright (C) 2010 Benjamin Moody
;;;
;;; This program is free software: you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;; GetFirstApp:
;;
;; Determine the starting page for the first Flash app.
;;
;; Output:
;; - A = first starting page (may or may not be an actual app)
;;
;; Destroys:
;; - F

GetFirstApp:
	in	a, (2)
	add	a, a
	ld	a, 15h
	ret	nc
	call	GetFlashPageMask
	and	69h
	ret

;; CheckUserFlashPage:
;;
;; Check if page A is within the "user Flash" area.
;;
;; Input:
;; - A = possible app starting page
;;
;; Output:
;; - Carry flag clear if A is a valid app or archive page
;; - Carry flag set if page is a RAM, OS, or reserved page
;;
;; Destroys:
;; - F

CheckUserFlashPage:
	cp	8
	ret	c
	push	bc
	 ld	b, a
	 call	GetFirstApp
	 add	a, 2
	 cp	b
	 ld	a, b
	 pop	bc
	ret

;; GetNextApp:
;;
;; Move to the next Flash application.
;;
;; Input:
;; - A = Flash app starting page
;;
;; Output:
;; - Carry flag set if no more (valid) apps
;; - A = port 6 = (appStartPage) = next app starting page
;; - (appNumPages) = number of pages in the next app
;;
;; Destroys:
;; - F, BC

GetNextApp:
	call	SelectApp
	ret	c
	ld	c, a
	ld	a, (appNumPages)
	ld	b, a
	ld	a, c
	sub	b
SelectApp:
	call	CheckUserFlashPage
	ret	c
	push	hl
	 push	de
	  ld	(appStartPage), a
	  out	(6), a
	  ld	b, a
	  ld	hl, (4000h)
	  ld	a, l
	  and	7Fh
	  jr	nz, SelectApp_Failed
	  ld	a, h
	  xor	0Fh
	  jr	nz, SelectApp_Failed
	  ld	a, b
	  BCALL	_FindAppNumPages
	  ld	a, b
	  sub	c
	  cp	b
	  jr	nc, SelectApp_Failed
	  ld	a, c
	  ld	(appNumPages),a
	  scf
SelectApp_Failed:
	  ccf
	  pop	de
	 pop	hl
	ld	a, b
	ret

;; GetPrevApp:
;;
;; Move to the previous Flash application.
;;
;; Input:
;; - A = Flash app starting page
;;
;; Output:
;; - Carry flag set if no more (valid) apps
;; - A = port 6 = (appStartPage) = previous app starting page
;; - (appNumPages) = number of pages in the next app
;;
;; Destroys:
;; - F, BC

GetPrevApp:
	inc	a
	ld	c, a
	call	GetFirstApp
	cp	c
	ret	c
	BCALL	_GetBaseVer
	or	a
	jr	z, GetPrevApp_BABT
	dec	a
	jr	nz, GetPrevApp_System
	ld	a, b
	cp	13
	jr	nc, GetPrevApp_System
GetPrevApp_BABT:
	ld	a, c
	add	a, low(baseAppBrTab)
	ld	c, a
	ld	b, high(baseAppBrTab)
	ld	a, (bc)
	jr	SelectApp

GetPrevApp_System:
	ld	a, c
	push	hl
	 push	de
	  BCALL _GetAppBasePage
	  pop	de
	 pop	hl
	jr	SelectApp

;; GetAppName:
;;
;; Retrieve app's name as a zero-terminated string.  (If app does not
;; have a defined name, return the string "?".)
;;
;; Input:
;; - A = starting page
;;
;; Output:
;; - Port 6 = app starting page
;; - HL = address of a zero-terminated string, in RAM (normally stored
;;   in OP1)
;;
;; Destroys:
;; - AF, BC, DE, HL

GetAppName:
	out	(6), a
	ld	hl, 4000h
	ld	de, 8040h
	BCALL	_FindAppHeaderSubField
	jr	nz, GetAppName_Unknown
	in	a, (6)
	inc	hl
	BCALL	_GetFieldSize
	inc	bc
	ld	de, OP1
	push	de
	 ldir
	 dec	de
	 xor	a
	 ld	(de), a
	 pop	hl
	ret

GetAppName_Unknown:
	ld	hl, unknownName
	ret

;; GetAppSize:
;;
;; Get the length of an application (the 80 0F field at the start) and
;; check that it's reasonable.  (The app length must be set correctly
;; before signing.)
;;
;; Input:
;; - A = app starting page
;;
;; Output:
;; - Zero flag clear if length is not valid
;; - Port 6 = (appStartPage) = app starting page
;; - (appNumPages) = number of pages in the app
;; - (appFinalPageLength) = total length of app data on last page
;;   (i.e., address of signature minus 4000h)
;;
;; Destroys:
;; - AF, BC, DE, HL, IX

GetAppSize:
	call	SelectApp
	;; check that length is
	;; - at least 64
	;; - at least 16384 * (npages - 1)
	;; - at most (16384 * npages) - 69
	ld	hl, (4002h)
	ld	de, (4004h)
	ld	a, d
	add	a, 6
	ld	d, a
	ld	a, e
	adc	a, l		; L should be zero
	ld	e, a
	ld	a, h
	adc	a, l
	ld	h, a		; LHED = length (including first 6
				; bytes)

	ld	a, h
	and	0C0h
	or	l
	ret	nz

	ld	l, e
	add	hl, hl
	add	hl, hl
	ld	a, (appNumPages)
	dec	a
	cp	h
	ret	nz
	or	a
	jr	nz, GetAppSize_Multipage
	ld	a, d
	and	0C0h
	or	e
	jr	z, GetAppSize_Error
GetAppSize_Multipage:
	ld	l, d
	ld	a, e
	and	3Fh
	ld	h, a
	ld	(appFinalPageLength), hl
	xor	a
	ld	de, 69 - 4000h
	add	hl, de
	ret	nc
GetAppSize_Error:
	or	1
	ret

