;;; -*- TI-Asm -*-

;;; Example code for installing a single-page app

	processor z80

	nolist
	include <ti83plus.inc>
	list

appCacheFlag	equ 9C87h

_IsInSet                         equ 44E9h
_FlashWriteDisable               equ 4F3Ch
_FillBasePageTable               equ 5011h
_GetVariableData                 equ 5092h
_WriteAByteUnsafe                equ 8021h
_LoadAIndPaged                   equ 8051h
_GetCertificateStart             equ 8057h
_WriteFlashUnsafe                equ 8087h
_WriteAByteSafe                  equ 80C6h
_WriteFlashSafe                  equ 80C9h
_SetAppLimit                     equ 80CFh

BCALL macro addr
	rst 28h
	;; UNREACHABLE
	dw addr
	endm


	org	userMem - 2
	db	0BBh, 6Dh

;;; Check if batteries are good.

;;; (If power fails in the middle of the operation, the user could
;;; potentially be left with a broken application and no way to
;;; uninstall it short of a full archive reset.)

	BCALL	_Chk_Batt_Low
	ld	hl, BatteryLowStr
	jp	z, DispError

;;; Check if app already exists.

;;; (If two apps on the calculator have the same name, only the first
;;; one can actually be used; there may also be more subtle problems
;;; with the OS and other programs that expect app names to be
;;; unique.)

	ld	hl, MyAppName - 1
	rst	rMOV9TOOP1
	ld	a, AppObj
	ld	(OP1), a
	BCALL	_FindApp
	ld	hl, AlreadyExistsStr
	jp	nc, DispError

;;; Find the first free application page.

	call	GetFirstApp
	ld	b, a
FindLastApp_Loop:
	ld	hl, 4000h
	BCALL	_LoadDEIndPaged
	ld	a, e
	and	7Fh
	jr	nz, FindLastApp_Done
	ld	a, d
	cp	0Fh
	jr	nz, FindLastApp_Done
	ld	a, b
	BCALL	_FindAppNumPages
	ld	a, b
	sub	c
	cp	b
	ld	b, a
	jr	c, FindLastApp_Loop
FindLastApp_Error:
	ld	hl, InvalidAppStr
	jp	DispError
FindLastApp_Done:

;;; Check that the page does actually begin with FF FF.

	ld	a, d
	and	e
	inc	a
	jr	nz, OutOfSpace

;;; Check that the page number is reasonable and that the sector is
;;; not already in use.

;;; (If we try to install an application in an existing archive or
;;; temporary sector, any number of really bad things might happen.)

	ld	a, b
	cp	12
	jr	c, OutOfSpace
	and	~3
	ld	hl, 4000h
	BCALL	_LoadAIndPaged
	inc	a
OutOfSpace:
	ld	hl, OutOfSpaceStr
	jr	nz, DispError

;;; Write our new application.

	push	bc
	 call	Unlock
	 pop	bc
	ld	hl, UnlockingFailedStr
	jr	c, DispError

	push	bc
	 ld	a, b
	 ld	hl, MyAppData
	 ld	de, 4000h
	 ld	bc, MyAppDataEnd - MyAppData
	 BCALL	_WriteFlashSafe
	 pop	bc
	ld	hl, WritingFailedStr
	jr	nz, DispError

;;; Clear app installed bit.

;;; (Often this bit will already be clear.  If the bit isn't clear,
;;; the OS will assume the app is garbage - perhaps left over from a
;;; failed linking or defragmenting operation - and delete it at the
;;; next opportunity.)

	call	GetFirstApp
	sub	b
	ld	c, a		; C = index of app page
	push	bc
	 BCALL	_GetCertificateStart
	 push	hl
	  ld	de, 1FE0h
	  add	hl, de
	  ld	a, c
	  inc	a
	  BCALL	_IsInSet	; HL = address, A = bit to clear
	  cpl
	  ld	d, a
	  call	GetFlashPageMask
	  dec	a		; A = certificate page
	  push	af
	   BCALL _LoadAIndPaged
	   and	d
	   ld	b, a
	   pop	af
	  ex	de, hl
	  BCALL _WriteAByteUnsafe ; (ignore errors)
	  pop	hl
	 pop	bc

;;; Set expiration count to infinite.

;;; (If we don't do this, the OS assumes the app is supposed to expire
;;; after being run 16 times.)

	push	bc
	 ld	b, 0
	 add	hl, bc
	 add	hl, bc
	 in	a, (2)
	 add	a, a
	 ld	de, 1F18h
	 jr	nc, SetExpCount_BE
	 ld	de, 1E50h
SetExpCount_BE:
	 add	hl, de
	 ex	de, hl		; DE = address of expiration count
	 BCALL	_OP1Set0
	 ld	hl, OP1 + 1	; (HL) = 80h, (HL+1) = 0
	 call	GetFlashPageMask
	 dec	a		; A = certificate page
	 ld	c, 2
	 BCALL	_WriteFlashUnsafe ; (ignore errors)

;;; Mark installed-app cache invalid.

;;; (If we don't do this, newer OSes may not show the app in the Apps
;;; menu.)

	 xor	a
	 ld	(appCacheFlag), a

;;; Set execution limit.

;;; (If we don't do this, the calculator will crash when attempting to
;;; execute the new app.)

	 pop	af
	dec	a
	BCALL	_SetAppLimit

;;; Rebuild base page table.

;;; (Usually unneeded for single-page apps, but a good idea anyway.
;;; If we don't do this, internal B_CALLs won't work.)

	BCALL	_FillBasePageTable

;;; We're done (finally!)

	ld	hl, SuccessStr
	jr	Done

DispError:
	push	hl
	 ld	hl, ErrorStr
	 BCALL	_PutS
	 BCALL	_NewLine
	 pop	hl
Done:
	BCALL	_PutS
	BCALL	_NewLine
	BCALL	_FlashWriteDisable
	ret


;; GetFirstApp:
;;
;; Determine the starting page for the first Flash app.
;;
;; Output:
;; - A = first starting page (may or may not be an actual app)
;;
;; Destroys:
;; - F

GetFirstApp:
	in	a, (2)
	add	a, a
	ld	a, 15h
	ret	nc
	call	GetFlashPageMask
	and	69h
	ret

;; GetFlashPageMask:
;;
;; Determine the last Flash page, or the mask of bits used for Flash
;; pages, for this calculator model.
;;
;; Output:
;; - A = last Flash page
;;
;; Destroys:
;; - F

GetFlashPageMask:
	in	a, (2)
	add	a, a
	ld	a, 1Fh
	ret	nc
	in	a, (21h)
	and	3
	ld	a, 7Fh
	ret	nz
	rra
	ret

;; ...

Unlock:
	di
	push iy
	 in a, (6)
	 ld c, a
	 ld de, 7B09h
	 ld a, e
	 ld (arcInfo), a
	 out (6), a
	 ld a, (de)
	 ld b, a
	 push bc
	  push de
	   call SetFlashPageD
	   ld a, e
	   B_CALL _WriteAByteSafe
	   ld hl, Unlock_Recover - 2
	   ld de, pagedBuf - 2
	   B_CALL _Mov18B
	   ld (pagedBuf + 2), sp
	   call Unlock_Main
	   pop af
	  pop bc
Unlock_Done:
	 pop iy
	ld a, c
	jr SetPage

SetFlashPageD:
	call GetFlashPageMask
	and d
SetPage:
	out (6), a
	ret

Unlock_Main:
	ld a, (_GetVariableData + 2)
	ld d, a
	ld hl, ramCode
	ld bc, 304
Unlock_Loop1:
	ld a, 0EDh
	cpir
	ret nz
	ld a, (hl)
	sub 0A0h
	jr nz, Unlock_Loop1
	ld c, (hl)
	ld b, a
	inc hl
	ld (iMathPtr5), hl
	ld hl, (_GetVariableData)
	inc a
Unlock_Data:
	ld (pagedCount), a
	or a
	call nz, SetFlashPageD
Unlock_Loop2:
	ld a, 3Ah
	cpir
	ret nz
	ld de, Unlock_Data + 1
Unlock_Loop3:
	ld a,(de)
	cp (hl)
	jr nz, Unlock_Loop2
	inc hl
	inc de
	cp 0C4h
	jr nz, Unlock_Loop3
	B_CALL _LdHLind
	ld iy, 56h - 25h
Unlock_Recover:
	jp (hl)
	   ld sp, 0
	   pop hl
	  pop bc
	 ld a, b
	 ld (hl), a
Unlock_WriteByte:
	 cp (hl)
	 jr nz, Unlock_WriteByte
	 ld (hl), 0F0h
	 jp Unlock_Done

;;; Example application

MyAppData:
	db 80h, 0Fh, 0, 0, 0, MyAppEnd - MyAppStart
MyAppStart:
	 db 80h, 12h
	  db 01h, 04h
	 db 80h, 21h
	  db 1
	 db 80h, 31h
	  db 1
	 db 80h, 48h
MyAppName:
	  db "SayHello"
	 db 80h, 81h
	  db 1
	 db 80h, 90h
	 db 80h, 0A1h
	  db 0FFh
	 db 03h, 26h
	  db 09h, 04h
	   db 04h, 6Fh, 1Bh, 80h
	 db 02h, 0Dh, 40h
	  db 0A1h, 06Bh, 099h, 0F6h, 059h, 0BCh, 067h, 0F5h
	  db 085h, 09Ch, 009h, 06Ch, 00Fh, 0B4h, 003h, 09Bh
	  db 0C9h, 003h, 032h, 02Ch, 0E0h, 003h, 020h, 0E3h
	  db 02Ch, 0F4h, 02Dh, 073h, 0B4h, 027h, 0C4h, 0A0h
	  db 072h, 054h, 0B9h, 0EAh, 07Ch, 03Bh, 0AAh, 016h
	  db 0F6h, 077h, 083h, 07Ah, 0EEh, 01Ah, 0D4h, 042h
	  db 04Ch, 06Bh, 08Bh, 013h, 01Fh, 0BBh, 093h, 08Bh
	  db 0FCh, 019h, 01Ch, 03Ch, 0ECh, 04Dh, 0E5h, 075h
	 db 80h, 7Fh, 0, 0, 0, 0

	BCALL	_ClrLCDFull
	ld	hl, 0
	ld	(curRow), hl
	ld	hl, String + 4000h - MyAppData
	ld	de, OP1
	push	de
	 BCALL	_StrCopy
	 pop	hl
	BCALL	_PutS
	BCALL	_GetKey
	BCALL	_JForceCmdNoChar
	;; UNREACHABLE
String:
	db	"Hello world", 0

MyAppEnd:
	db	02h, 2Dh, 40h
MyAppDataEnd:


SuccessStr:
	db	"Success", 0
ErrorStr:
	db	"ERROR", 0
BatteryLowStr:
	db	"Battery low", 0
AlreadyExistsStr:
	db	"App already     exists", 0
InvalidAppStr:
	db	"Invalid app(s)  installed", 0
OutOfSpaceStr:
	db	"Not enough free archive mem", 0
UnlockingFailedStr:
	db	"Cannot unlock   Flash", 0
WritingFailedStr:
	db	"Error writing toFlash", 0
