;;; -*- TI-Asm -*-
;;;
;;; DuckSign - sign Flash applications on the TI-83 Plus
;;;
;;; Copyright (C) 2010 Benjamin Moody
;;;
;;; This program is free software: you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
	
;; HashCurApp:
;;
;; Compute MD5 hash of the current application.
;;
;; Input:
;; - (appStartPage) = starting page
;; - (appNumPages) = number of pages
;; - (appFinalPageLength) = length of final page
;; - (penRow) = Y position to display progress bar (set to -1 to
;;   disable progress bar)
;;
;; Output:
;; - Carry flag set if ON key was pressed (calculation aborted)
;; - (MD5Hash) = hash of application (16 bytes)
;;
;; Destroys:
;; - AF, BC, DE, HL, IX
;; - Port 6
;; - appData
;; - MD5Buffer, MD5AltRegs

HashCurApp:
	ld	a, (appStartPage)
	out	(6), a

	ld	hl, (appFinalPageLength)
	add	hl, hl
	add	hl, hl
	ld	a,(appNumPages)
	dec	a
	ld	l, h
	ld	h, a
	inc	hl
	call	InitProgress

  ifdef INTERNAL_MD5
	call	MD5Init
  else
	BCALL	_MD5Init
  endif

	ld	a, (appNumPages)
	ld	c, a
	jr	HashApp_Begin

HashApp_PageLoop:
	ld	hl, 4000h
	ld	b, l
HashApp_BlockLoop:
	call	CheckOnAbort
	ret	c
	call	HashBlock
	djnz	HashApp_BlockLoop
	in	a, (6)
	dec	a
	out	(6),a
HashApp_Begin:
	dec	c
	jr	nz, HashApp_PageLoop

	ld	hl, (appFinalPageLength)
	add	hl, hl
	add	hl, hl
	ld	b, h
	ld	hl, 4000h
	inc	b
	jr	HashApp_LastPageBegin

HashApp_LastPageLoop:
	call	CheckOnAbort
	ret	c
	call	HashBlock
HashApp_LastPageBegin:
	djnz	HashApp_LastPageLoop

	ld	a, (appFinalPageLength)
	and	63
	ld	c, a
	ld	b, 0
	call	nz, HashPartial

  ifdef INTERNAL_MD5
	call	MD5Final
  else
	BCALL	_MD5Final
  endif
	jp	UpdateProgress


;; HashBlock:
;;
;; Hash a 64-byte block of data.
;;
;; Input:
;; - (MD5Length) = length of data hashed so far
;; - (MD5Hash) = current MD5 registers
;; - HL = address of data
;;
;; Output:
;; - (MD5Length), (MD5Hash) updated
;; - HL = input HL plus 64
;;
;; Destroys:
;; - AF, DE, IX
;; - appData
;; - MD5Buffer

HashBlock:
	push	bc
	 ld	bc, 64
	 call	HashPartial
	 push	hl
	  call	UpdateProgress
	  pop	hl
	 pop	bc
	ret

;; HashPartial:
;;
;; Queue a block of data to be hashed.
;;
;; Input:
;; - (MD5Length) = length of data hashed so far
;; - (MD5Hash) = current MD5 registers
;; - HL = address of data
;; - BC = length of data
;;
;; Output:
;; - (MD5Length), (MD5Hash) updated
;; - HL = input HL plus BC
;;
;; Destroys:
;; - AF, BC, DE, IX
;; - appData
;; - MD5Buffer

HashPartial:
  ifdef INTERNAL_MD5
	push	bc
	 call	MD5Update
	 pop	bc
  else
	ld	de, appData
	push	de
	 push	bc
	  ldir
	  pop	bc
	 ex	(sp), hl
	 BCALL	_MD5Update
	 pop	hl
  endif
	ret
