;;; -*- TI-Asm -*-
;;;
;;; DuckSign - sign Flash applications on the TI-83 Plus
;;;
;;; Copyright (C) 2010 Benjamin Moody
;;;
;;; This program is free software: you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

  ifdef INTERNAL_MD5

MD5_A	equ MD5Hash
MD5_B	equ MD5Hash + 4
MD5_C	equ MD5Hash + 8
MD5_D	equ MD5Hash + 12

MD5_AA	equ MD5AltRegs
MD5_BB	equ MD5AltRegs + 4
MD5_CC	equ MD5AltRegs + 8
MD5_DD	equ MD5AltRegs + 12

;; MD5Init:
;;
;; Reset the MD5 registers for a new computation.
;;
;; Output:
;; - (MD5Length) = 0
;; - MD5Hash set to initial values
;;
;; Destroys:
;; - None

MD5Init:
	push	hl
	 push	de
	  push	bc
	   ld	hl, MD5Length
	   ld	b, 8
	   BCALL _ClrLp
	   ld	hl, md5InitialRegisters
	   ld	de, MD5Hash
	   call	Mov16B
	   pop	bc
	  pop	de
	 pop	hl
	ret

md5InitialRegisters:
	db 001h, 023h, 045h, 067h
	db 089h, 0ABh, 0CDh, 0EFh
	db 0FEh, 0DCh, 0BAh, 098h
	db 076h, 054h, 032h, 010h

;; MD5Update:
;;
;; Add a chunk of data to the MD5 hash.
;;
;; Input:
;; - (MD5Length) = length of data hashed so far
;; - MD5Hash contains current register values
;; - MD5Buffer contains queued data
;; - HL = address of data to hash
;; - BC = number of bytes to hash
;;
;; Output:
;; - (MD5Length), MD5Hash, MD5Buffer updated
;; - HL = input HL plus BC (note that this is not done by the boot
;;   code version of MD5Update)
;;
;; Destroys:
;; - AF, BC, DE, IX
;; - MD5AltRegs

MD5Update:
	ld	a, b
	or	c
	ret	z
	ld	de, (MD5Length)

	;; we want de = (length/8 mod 64) + MD5_Buffer
	;; and a = length/2 mod 256
	srl	d
	ld	a, e
	rra
	ld	d, a
	rrca
	rrca
	add	a, low(MD5Buffer)
	ld	e, a
	ld	a, d
	ld	d, high(MD5Buffer)
MD5Update_Loop:
	add	a,4		; add one byte
	ldi
	jr	c, MD5Update_Process	; if the addition carried, the
					; length now = 0 mod 64
	jp	pe, MD5Update_Loop	; if parity is clear (odd), BC
					; now = 0
MD5Update_SetLength:
	ld	de, MD5Length
	add	a, a
	ld	(de), a
	inc	de
	ld	a, (de)
	rra
	rlca
	ld	(de), a
	ret

MD5Update_Process:
	push	hl
	 push	bc

	  ;; A = bits 1-9 of the MD5 length
	  ;; (minus 256 = 512 bits = 64 bytes, which carried)
	  call	MD5Update_SetLength
	  ex	de, hl
	  add	a, 2
	  ld	(hl), a
	  call	c, CarryInc

	  ;; Now process the current 64-byte block
	  ld	hl, MD5Hash
	  ld	de, MD5AltRegs
	  call	Mov16B

	  call	MD5_FourRounds

	  ld	hl, MD5_AA
	  ld	de, MD5_A
	  ld	b, 4
MD5Update_AddLoop:
	  ld	a, (de)
	  add	a, (hl)
	  ld	(de), a
	  inc	l
	  inc	e
	  ld	a, (de)
	  adc	a, (hl)
	  ld	(de), a
	  inc	l
	  inc	e
	  ld	a, (de)
	  adc	a, (hl)
	  ld	(de), a
	  inc	l
	  inc	e
	  ld	a, (de)
	  adc	a, (hl)
	  ld	(de), a
	  inc	l
	  inc	e
	  djnz	MD5Update_AddLoop
	  pop	bc
	 pop	hl
	jr	MD5Update

;; MD5Final:
;;
;; Complete the MD5 hash.
;;
;; Input:
;; - (MD5Length) = length of data hashed so far
;; - MD5Hash contains current register values
;; - MD5Buffer contains queued data
;;
;; Output:
;; - MD5Hash = completed hash
;; - (MD5Hash - 1) = length of hash as a a bignum
;;
;; Destroys:
;; - AF, BC, DE, HL, IX
;; - MD5Length
;; - MD5Buffer
;; - MD5AltRegs
;; - MD5Temp

MD5Final:
	ld	hl, MD5Length
	ld	de, MD5Temp
	BCALL	_Mov8B

	;; add 80 byte
	ld	hl, md5PadData
	ld	bc, 1
	call	MD5Update
	;; and as many 00 bytes as needed to get 56 mod 64
	inc	hl
MD5Final_PadLoop:
	ld	a, (MD5Length)
	cp	0c0h
	jr	z, MD5Final_24mod32
MD5Final_ContinuePadding:
	dec	hl
	inc	c
	call	MD5Update
	jr	MD5Final_PadLoop

MD5Final_24mod32:
	ld	a, (MD5Length + 1)
	and	1
	jr	z, MD5Final_ContinuePadding

	ld	hl, MD5Temp
	ld	bc, 8
	call	MD5Update

	ld	hl, MD5Hash - 1
	ld	(hl), 16
	jp	ShortenBignum

md5PadData:
	db 80h, 00h

;; MD5_FourRounds:
;;
;; Perform the four rounds of the main MD5 calculation.
;;
;; Input:
;; - MD5Hash contains register values
;; - MD5Buffer contains data to be hashed
;;
;; Output:
;; - MD5Hash contains new register values
;;
;; Destroys:
;; - AF, BC, DE, HL, IX

MD5_FourRounds:
	ld	hl, md5Table
	in	a, (2)
	add	a, a
	jr	c, MD5_FourRounds_Hardware

	ld	b, 16
	ld	ix, md5RegsTable
MD5_FourRounds_Loop1:
	call	MD5_FF
	djnz	MD5_FourRounds_Loop1
	ld	b, 16
	ld	ix, md5RegsTable
MD5_FourRounds_Loop2:
	call	MD5_GG
	djnz	MD5_FourRounds_Loop2
	ld	b, 16
	ld	ix, md5RegsTable
MD5_FourRounds_Loop3:
	call	MD5_HH
	djnz	MD5_FourRounds_Loop3
	ld	b, 16
	ld	ix, md5RegsTable
MD5_FourRounds_Loop4:
	call	MD5_II
	djnz	MD5_FourRounds_Loop4
	ret

MD5_FourRounds_Hardware:
	xor	a
MD5_FourRounds_HardwareLoop:
	out	(1Fh), a
	ld	c, a
	ld	b, 16
	ld	ix, md5RegsTable
MD5_FourRounds_HardwareSubLoop:
	call	MD5_HardwareOp
	djnz	MD5_FourRounds_HardwareSubLoop
	ld	a, c
	inc	a
	cp	4
	jr	c, MD5_FourRounds_HardwareLoop
	ret

;; MD5_HardwareOp:
;;
;; Perform a single MD5 operation (FF, GG, HH, or II) using the
;; hardware ports available on the 83+ SE and 84+.
;;
;; Input:
;; - Appropriate operation selected in port 1F
;; - MD5Hash contains register values
;; - MD5Buffer contains data to be hashed
;; - HL = address of entry in md5Table (specifying which values to
;;   use)
;; - IX = address in md5RegsTable (specifying which MD5 registers to
;;   use)
;;
;; Output:
;; - MD5Regs updated
;; - HL = address of next entry in md5Table
;; - IX = address of next entry in md5RegsTable
;;
;; Destroys:
;; - AF, DE

MD5_HardwareOp:
	push	bc
	 push	hl
	  ld	l, (ix + 0)
	  ld	h, high(MD5Hash)
	  ld	c, 18h		; port 18h: initial a
	  call	Out32

	  ld	l, (ix + 1)
	  inc	c		; port 19h: b
	  call	Out32

	  ld	l, (ix + 2)
	  inc	c		; port 1ah: c
	  call	Out32

	  ld	l, (ix + 3)
	  inc	c		; port 1bh: d
	  call	Out32

	  pop	hl
	 push	hl
	  ld	l, (hl)
	  ld	h, high(MD5Buffer)
	  inc	c		; port 1ch: X (or T)
	  call	Out32

	  pop	hl
	 inc	hl
	 inc	c		; port 1dh: T (or X)
	 call	Out32

	 inc	c		; port 1eh: shift
	 outi
	 push	hl

	  ;; Store the result back into a, and advance to the next
	  ;; entry in the md5RegsTable.

	  ld	l, (ix + 0)
	  ld	h, high(MD5Hash)
	  ld	c, 1Ch
	  ini
	  inc	c
	  ini
	  inc	c
	  ini
	  inc	c
	  ini

	  dec	ix
	  pop	hl
	 pop	bc
	ret

;; Out32:
;;
;; Write a 32-bit number to a hardware port, LSB first.
;;
;; Input:
;; - HL = address of number
;; - C = port address
;;
;; Output:
;; - HL = input HL plus 4
;;
;; Destroys:
;; - F, B

Out32:
	outi
	outi
	outi
	outi
	ret

;; MD5_FF:
;;
;; Perform an FF operation.
;;
;; Input:
;; - MD5Hash contains register values
;; - MD5Buffer contains data to be hashed
;; - HL = address of entry in md5Table (specifying which values to
;;   use)
;; - IX = address in md5RegsTable (specifying which MD5 registers to
;;   use)
;;
;; Output:
;; - MD5Regs updated
;; - HL = address of next entry in md5Table
;; - IX = address of next entry in md5RegsTable
;;
;; Destroys:
;; - AF, DE

MD5_FF:
	push	bc
	 push	hl

	  ;; Compute FF(a,b,c,d,k,s,t)
	  ;; IX -> <a <b <c <d
	  ;; HL -> (<MD5_Buffer + 4k) <<t ><t <>t >>t s
	  ;;
	  ;; FF(a,b,c,d,k,s,t) = b + ((a + F(b,c,d) + X[k] + t) <<< s)
	  ;;
	  ;; where  F(X,Y,Z) = XY v not(X) Z
	  ;;                 = Z xor X (Y xor Z)

	  ld	l, (ix + 2)
	  ld	h, high(MD5Hash)
	  call	Load_BCDE_iHL	; c

	  ld	l, (ix + 3)
	  call	Xor_BCDE_iHL	; c xor d

	  ld	l, (ix + 1)
	  call	And_BCDE_iHL	; b (c xor d)

	  ld	l, (ix + 3)
	  call	Xor_BCDE_iHL	; d xor b (c xor d) = F(b,c,d)

MD5_Operation_Main:
	  ;; The rest of the computation is the same for all four
	  ;; rounds.

	  ld	l, (ix + 0)
	  call	Add_BCDE_iHL	; a + f(b,c,d)

	  pop	hl
	 push	hl
	  ld	l, (hl)
	  ld	h, high(MD5Buffer)
	  call	Add_BCDE_iHL	; a + f(b,c,d) + X[k]

	  pop	hl
	 inc	hl
	 call	Add_BCDE_iHL	; a + f(b,c,d) + X[k] + t
	 inc	hl
	 ld	a, (hl)
	 inc	hl		; advance to next MD5_Table entry
	 push	hl
	  call	RLC_BCDE_A	; (a + f(b,c,d) + X[k] + t) <<< s

	  ld	l, (ix + 1)
	  ld	h, high(MD5Hash)
	  call	Add_BCDE_iHL	; b + ((a + f(b,c,d) + X[k] + t) <<< s)

	  ;; We store the result back into a, and advance to the next
	  ;; entry in the MD5_RegsTable.

	  ld	l, (ix + 0)
	  ld	(hl), e
	  inc	hl
	  ld	(hl), d
	  inc	hl
	  ld	(hl), c
	  inc	hl
	  ld	(hl), b

	  dec	ix
	  pop	hl
	 pop	bc
	ret

;; MD5_GG:
;;
;; Perform a GG operation.
;;
;; Input:
;; - MD5Hash contains register values
;; - MD5Buffer contains data to be hashed
;; - HL = address of entry in md5Table (specifying which values to
;;   use)
;; - IX = address in md5RegsTable (specifying which MD5 registers to
;;   use)
;;
;; Output:
;; - MD5Regs updated
;; - HL = address of next entry in md5Table
;; - IX = address of next entry in md5RegsTable
;;
;; Destroys:
;; - AF, DE

MD5_GG:
	push	bc
	 push	hl

	  ;; Compute GG(a,b,c,d,k,s,t)
	  ;;
	  ;; GG(a,b,c,d,k,s,t) = b + ((a + G(b,c,d) + X[k] + t) <<< s)
	  ;;
	  ;; where  G(X,Y,Z) = XZ v Y not(Z)
	  ;;                 = Y xor Z (X xor Y)

	  ld	l, (ix + 1)
	  ld	h, high(MD5Hash)
	  call	Load_BCDE_iHL	; b

	  ld	l, (ix + 2)
	  call	Xor_BCDE_iHL	; b xor c

	  ld	l, (ix + 3)
	  call	And_BCDE_iHL	; d (b xor c)

	  ld	l, (ix + 2)
	  call	Xor_BCDE_iHL	; c xor d (b xor c) = G(b,c,d)

	  jr	MD5_Operation_Main

;; MD5_HH:
;;
;; Perform an HH operation.
;;
;; Input:
;; - MD5Hash contains register values
;; - MD5Buffer contains data to be hashed
;; - HL = address of entry in md5Table (specifying which values to
;;   use)
;; - IX = address in md5RegsTable (specifying which MD5 registers to
;;   use)
;;
;; Output:
;; - MD5Regs updated
;; - HL = address of next entry in md5Table
;; - IX = address of next entry in md5RegsTable
;;
;; Destroys:
;; - AF, DE

MD5_HH:
	push	bc
	 push	hl

	  ;; Compute HH(a,b,c,d,k,s,t)
	  ;;
	  ;; HH(a,b,c,d,k,s,t) = b + ((a + H(b,c,d) + X[k] + t) <<< s)
	  ;;
	  ;; where  H(X,Y,Z) = X xor Y xor Z

	  ld	l, (ix + 1)
	  ld	h, high(MD5Hash)
	  call	Load_BCDE_iHL	; b

	  ld	l, (ix + 2)
	  call	Xor_BCDE_iHL	; b xor c

	  ld	l, (ix + 3)
	  call	Xor_BCDE_iHL	; b xor c xor d = H(b,c,d)

	  jr	MD5_Operation_Main

;; MD5_II
;;
;; Perform an II operation.
;;
;; Input:
;; - MD5Hash contains register values
;; - MD5Buffer contains data to be hashed
;; - HL = address of entry in md5Table (specifying which values to
;;   use)
;; - IX = address in md5RegsTable (specifying which MD5 registers to
;;   use)
;;
;; Output:
;; - MD5Regs updated
;; - HL = address of next entry in md5Table
;; - IX = address of next entry in md5RegsTable
;;
;; Destroys:
;; - AF, DE

MD5_II:
	push	bc
	 push	hl

	  ;; Compute II(a,b,c,d,k,s,t)
	  ;;
	  ;; II(a,b,c,d,k,s,t) = b + ((a + I(b,c,d) + X[k] + t) <<< s)
	  ;;
	  ;; where  I(X,Y,Z) = Y xor (X v not(Z))

	  ld	l, (ix + 3)
	  ld	h, high(MD5Hash)
	  ld	a, (hl)
	  cpl
	  ld	e, a
	  inc	hl
	  ld	a, (hl)
	  cpl
	  ld	d, a
	  inc	hl
	  ld	a, (hl)
	  cpl
	  ld	c, a
	  inc	hl
	  ld	a, (hl)
	  cpl
	  ld	b, a		; not(d)

	  ld	l, (ix + 1)
	  call	Or_BCDE_iHL	; b v not(d)

	  ld	l, (ix + 2)
	  call	Xor_BCDE_iHL	; c xor (b v not(d)) = H(b,c,d)

	  jp	MD5_Operation_Main

;; Load_BCDE_iHL:
;;
;; Load BCDE from (HL) in little-endian order.
;;
;; Input:
;; - HL = address of a 32-bit integer
;;
;; Output:
;; - BCDE = (HL)
;; - HL increased by 3
;;
;; Destroys:
;; - None

Load_BCDE_iHL:
	ld	e, (hl)
	inc	hl
	ld	d, (hl)
	inc	hl
	ld	c, (hl)
	inc	hl
	ld	b, (hl)
	ret

;; And_BCDE_iHL:
;;
;; AND BCDE with (HL).
;;
;; Input:
;; - HL = address of a 32-bit integer
;; - BCDE = 32-bit integer
;;
;; Output:
;; - BCDE = BCDE & (HL)
;; - HL increased by 3
;;
;; Destroys:
;; - AF

And_BCDE_iHL:
	ld	a, (hl)
	and	e
	ld	e, a
	inc	hl
	ld	a, (hl)
	and	d
	ld	d, a
	inc	hl
	ld	a, (hl)
	and	c
	ld	c, a
	inc	hl
	ld	a, (hl)
	and	b
	ld	b, a
	ret

;; Or_BCDE_iHL:
;;
;; OR BCDE with (HL).
;;
;; Input:
;; - HL = address of a 32-bit integer
;; - BCDE = 32-bit integer
;;
;; Output:
;; - BCDE = BCDE v (HL)
;; - HL increased by 3
;;
;; Destroys:
;; - AF

Or_BCDE_iHL:
	ld	a, (hl)
	or	e
	ld	e, a
	inc	hl
	ld	a, (hl)
	or	d
	ld	d, a
	inc	hl
	ld	a, (hl)
	or	c
	ld	c, a
	inc	hl
	ld	a, (hl)
	or	b
	ld	b, a
	ret

;; Xor_BCDE_iHL:
;;
;; XOR BCDE with (HL).
;;
;; Input:
;; - HL = address of a 32-bit integer
;; - BCDE = 32-bit integer
;;
;; Output:
;; - BCDE = BCDE xor (HL)
;; - HL increased by 3
;;
;; Destroys:
;; - AF

Xor_BCDE_iHL:
	ld	a, (hl)
	xor	e
	ld	e, a
	inc	hl
	ld	a, (hl)
	xor	d
	ld	d, a
	inc	hl
	ld	a, (hl)
	xor	c
	ld	c, a
	inc	hl
	ld	a, (hl)
	xor	b
	ld	b, a
	ret

;; Add_BCDE_iHL:
;;
;; Add (HL) to BCDE.
;;
;; Input:
;; - HL = address of a 32-bit integer
;; - BCDE = 32-bit integer
;;
;; Output:
;; - BCDE = BCDE + (HL)
;; - HL increased by 3
;;
;; Destroys:
;; - AF

Add_BCDE_iHL:
	ld	a, (hl)
	add	a, e
	ld	e, a
	inc	hl
	ld	a, (hl)
	adc	a, d
	ld	d, a
	inc	hl
	ld	a, (hl)
	adc	a, c
	ld	c, a
	inc	hl
	ld	a, (hl)
	adc	a, b
	ld	b, a
	ret

;; RLC_BCDE_A:
;;
;; Rotate BCDE left circularly by A bits.
;;
;; Input:
;; - A = number of bits to rotate
;; - BCDE = 32-bit integer
;;
;; Output:
;; - BCDE = BCDE <<< A
;;
;; Destroys:
;; - AF, L

RLC_BCDE_A:
	sub	8
	jr	c, RLC_BCDE_A_Bits

	ld	l, b
	ld	b, c
	ld	c, d
	ld	d, e
	ld	e, l
	jr	RLC_BCDE_A

RLC_BCDE_A_Bits:
	add	a, 8
	ret	z
	ld	l, a
	ld	a, b
	ld	b, l
	ex	de, hl

RLC_BCDE_A_Loop:
	add	hl, hl
	rl	c
	rla
	jr	nc, RLC_BCDE_A_NC
	inc	l
RLC_BCDE_A_NC:
	djnz	RLC_BCDE_A_Loop

	ex	de, hl
	ld	b, a
	ret

;;; Table of offsets (k), shifts (s), and table constants (t) for each
;;; of the 64 MD5 operations

MTBL macro kkk, sss, tth, ttl
	db	(low(MD5Buffer)) + (4 * kkk)
	dw	ttl, tth
	db	sss
	endm

md5Table:
        MTBL  0,  7, 0d76ah,0a478h
        MTBL  1, 12, 0e8c7h,0b756h
        MTBL  2, 17, 02420h,070dbh
        MTBL  3, 22, 0c1bdh,0ceeeh
        MTBL  4,  7, 0f57ch,00fafh
        MTBL  5, 12, 04787h,0c62ah
        MTBL  6, 17, 0a830h,04613h
        MTBL  7, 22, 0fd46h,09501h
        MTBL  8,  7, 06980h,098d8h
        MTBL  9, 12, 08b44h,0f7afh
        MTBL 10, 17, 0ffffh,05bb1h
        MTBL 11, 22, 0895ch,0d7beh
        MTBL 12,  7, 06b90h,01122h
        MTBL 13, 12, 0fd98h,07193h
        MTBL 14, 17, 0a679h,0438eh
        MTBL 15, 22, 049b4h,00821h

        MTBL  1,  5, 0f61eh,02562h
        MTBL  6,  9, 0c040h,0b340h
        MTBL 11, 14, 0265eh,05a51h
        MTBL  0, 20, 0e9b6h,0c7aah
        MTBL  5,  5, 0d62fh,0105dh
        MTBL 10,  9, 00244h,01453h
        MTBL 15, 14, 0d8a1h,0e681h
        MTBL  4, 20, 0e7d3h,0fbc8h
        MTBL  9,  5, 021e1h,0cde6h
        MTBL 14,  9, 0c337h,007d6h
        MTBL  3, 14, 0f4d5h,00d87h
        MTBL  8, 20, 0455ah,014edh
        MTBL 13,  5, 0a9e3h,0e905h
        MTBL  2,  9, 0fcefh,0a3f8h
        MTBL  7, 14, 0676fh,002d9h
        MTBL 12, 20, 08d2ah,04c8ah

        MTBL  5,  4, 0fffah,03942h
        MTBL  8, 11, 08771h,0f681h
        MTBL 11, 16, 06d9dh,06122h
        MTBL 14, 23, 0fde5h,0380ch
        MTBL  1,  4, 0a4beh,0ea44h
        MTBL  4, 11, 04bdeh,0cfa9h
        MTBL  7, 16, 0f6bbh,04b60h
        MTBL 10, 23, 0bebfh,0bc70h
        MTBL 13,  4, 0289bh,07ec6h
        MTBL  0, 11, 0eaa1h,027fah
        MTBL  3, 16, 0d4efh,03085h
        MTBL  6, 23, 00488h,01d05h
        MTBL  9,  4, 0d9d4h,0d039h
        MTBL 12, 11, 0e6dbh,099e5h
        MTBL 15, 16, 01fa2h,07cf8h
        MTBL  2, 23, 0c4ach,05665h

        MTBL  0,  6, 0f429h,02244h
        MTBL  7, 10, 0432ah,0ff97h
        MTBL 14, 15, 0ab94h,023a7h
        MTBL  5, 21, 0fc93h,0a039h
        MTBL 12,  6, 0655bh,059c3h
        MTBL  3, 10, 08f0ch,0cc92h
        MTBL 10, 15, 0ffefh,0f47dh
        MTBL  1, 21, 08584h,05dd1h
        MTBL  8,  6, 06fa8h,07e4fh
        MTBL 15, 10, 0fe2ch,0e6e0h
        MTBL  6, 15, 0a301h,04314h
        MTBL 13, 21, 04e08h,011a1h
        MTBL  4,  6, 0f753h,07e82h
        MTBL 11, 10, 0bd3ah,0f235h
        MTBL  2, 15, 02ad7h,0d2bbh
        MTBL  9, 21, 0eb86h,0d391h

;;; Table of register addresses for each of the 16 operations in each
;;; round (we restart at the beginning of the next round.)

;;; This table is read backwards, since the sequence of registers
;;; shifts right after each instruction.  So after each operation we
;;; move back by one byte in this table.

	db            low MD5_B, low MD5_C, low MD5_D
	db low MD5_A, low MD5_B, low MD5_C, low MD5_D
	db low MD5_A, low MD5_B, low MD5_C, low MD5_D
	db low MD5_A, low MD5_B, low MD5_C, low MD5_D
md5RegsTable:
	db low MD5_A, low MD5_B, low MD5_C, low MD5_D

  endif
