;;; -*- TI-Asm -*-
;;;
;;; DuckSign - sign Flash applications on the TI-83 Plus
;;;
;;; Copyright (C) 2010 Benjamin Moody
;;;
;;; This program is free software: you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;; ReduceModStrict:
;;
;; Calculate the exact modular residue of a number, subtracting out
;; all multiples of the current modulus from the input.  The result
;; will be between zero and p-1, inclusive.
;;
;; (This routine is itself very inefficient, and should only be used
;; after calling ReduceModN.)
;;
;; Input:
;; - HL = address of value to be reduced (32-33 bytes, same length as
;;   modulus)
;; - Modulus values set up using SetupMod
;;
;; Output:
;; - Reduced value written back to the same buffer and zero padded
;;
;; Destroys:
;; - AF, BC, DE, BC', DE', HL'

ReduceModStrict:
	ex	de, hl
	ld	b, d
	ld	c, e
ReduceModStrict_Loop:
	call	SubModulus
	jr	nc, ReduceModStrict_Loop
	call	AddModulus
	ex	de, hl
	ret

;; SubtractMod:
;;
;; Compute the difference between two numbers in the current modulus.
;; Both must be in the approximate range of the modulus.  The result
;; may or may not be fully reduced depending on the inputs.
;;
;; Input:
;; - BC = first number (32-33 bytes, same length as modulus)
;; - HL = second number (32-33 bytes, same length as modulus)
;; - DE = buffer to place result
;; - Modulus values set up using SetupMod
;;
;; Output:
;; - Reduced value written to the given output buffer and zero padded
;;
;; Destroys:
;; - AF, BC, HL, BC', DE', HL'

SubtractMod:
	call	SubML
	ld	b, d
	ld	c, e
	ret	nc
SubtractMod_Loop:
	call	AddModulus
	jr	nc, SubtractMod_Loop
	ret

;; ReduceModN:
;;
;; Subtract out most (but not necessarily all) multiples of the
;; current modulus from the input number.  The result is guaranteed to
;; be less than 2^256 (if P is selected as the modulus) or less than
;; 2^264 (if Q is selected), and will be congruent to the original
;; number, but will occasionally still be greater than the modulus.
;;
;; Input:
;; - A = size of input, bytes
;; - HL = address of value to be reduced (A bytes, no carries)
;; - Modulus values set up using SetupMod
;;
;; Output:
;; - Reduced value written back to the same buffer and zero padded
;;
;; Destroys:
;; - AF, BC, DE, HL

ReduceModN:
	ld	c, a
	dec	a
	add	a, l
	FAIL_IF_C
	ld	l, a
	ex	de, hl
	jr	ReduceModN_Loop

ReduceModN_Sub:
	ld	b, (hl)
ReduceModN_Sub_Length:
	ld	a, e
	sub	b		; note: if the buffer begins on a
				; 256-byte boundary, this will
				; sometimes carry
	ld	e, a
	ld	a, c
	cp	b
	ret	c
ReduceModN_Sub_Loop:
	inc	e
	inc	l
	FAIL_IF_Z
	ld	a, (de)
	sbc	a, (hl)
	ld	(de), a
	djnz	ReduceModN_Sub_Loop
	jp	nz, ReduceModN_Loop
ReduceModN_DecLength:
	dec	c
	ret	z
	dec	e
	FAIL_IF_Z
ReduceModN_Loop:
	;; DE = address of MSB
	;; C = length of number

	;; Look at the MSB of the number.  The modJumpTable (which is
	;; pre-calculated by SetupMod) tells us, given the MSB, the
	;; largest power-of-two multiple of the modulus that we are
	;; guaranteed to be able to subtract out.

	;; By doing this, we avoid the need to restore the result if
	;; we guess incorrectly.  The drawbacks are (1) that we expect
	;; to require roughly one extra subtraction per byte, and (2)
	;; that this routine won't always reduce the number fully
	;; (although it's close enough for many purposes.)

	ld	a, (de)
	ld	l, a
	ld	h, high(modJumpTable)
	ld	l, (hl)
	ld	h, high(ReduceModN_DecLength)
	jp	(hl)

ReduceModN_7:
	ld	hl, modulusShift7
	jp	ReduceModN_Sub

ReduceModN_6:
	ld	hl, modulusShift6
	jp	ReduceModN_Sub

ReduceModN_5:
	ld	hl, modulusShift5
	jp	ReduceModN_Sub

ReduceModN_4:
	ld	hl, modulusShift4
	jp	ReduceModN_Sub

ReduceModN_3:
	ld	hl, modulusShift3
	jp	ReduceModN_Sub

ReduceModN_2:
	ld	hl, modulusShift2
	jp	ReduceModN_Sub

ReduceModN_1:
	ld	hl, modulusShift1
	jp	ReduceModN_Sub

ReduceModN_0:
	ld	hl, modulusShift0
	jp	ReduceModN_Sub

ReduceModN_minus1:
	ld	hl, modulusShift7
	ld	b, (hl)
	inc	b
	jp	ReduceModN_Sub_Length

ReduceModN_Broken:
	;; if we get here, something is seriously wrong
	im	1
	BCALL	_ErrDomain
	;; UNREACHABLE

	assert	high(ReduceModN_Broken) == high(ReduceModN_DecLength)

;; SetupMod:
;;
;; Set the modulus to be used by ReduceModN and ReduceModStrict.
;;
;; Input:
;; - HL = address of modulus (TI bignum format; must be odd)
;; - C = initial shift control value (80h left-rotated by the length
;;   of the modulus in base 2; or, the MSB of the modulus with all
;;   bits after the first 1 bit set to zero)
;;
;; Output:
;; - (modulusLength) = length of modulus
;; - (modulus) = address of modulus LSB
;;
;; - modJumpTable contains addresses to jump to in ReduceModN (see
;;   above)
;;
;; - modulusShift7 = modulus multiplied by the smallest power of two
;;   such that the total length is a multiple of 8 bits (i.e., the
;;   most significant byte has bit 7 set) (TI bignum format)
;;
;; - modulusShift6 = modulus multiplied by the smallest power of two
;;   such that the total length is 7 mod 8 (TI bignum format)
;;
;; ...
;;
;; - modulusShift0 = modulus multiplied by the smallest power of two
;;   such that the total length is 1 mod 8 (TI bignum format)
;;
;; Destroys:
;; - AF, BC, DE, HL

SetupMod_P0104:
	ld	hl, rabinP0104_length
	ld	c, P0104_INIT_BIT
	jr	SetupMod

SetupMod_Q0104:
	ld	hl, rabinQ0104_length
	ld	c, Q0104_INIT_BIT
SetupMod:
	ld	a, (hl)
	ld	(modulusLength), a
	inc	hl
	ld	(modulus), hl
	dec	hl

	;; Base cases: if MSB of a number to be reduced is zero, we
	;; skip it...
	ld	de, modJumpTable
	ld	a, low(ReduceModN_DecLength)
	ld	(de), a

	;; ...and if MSB is 1, we use the "bit -1" option (i.e.,
	;; subtract out modulusShift7, starting one byte earlier than
	;; usual)
	ld	a, low(ReduceModN_minus1)
	inc	e
	ld	(de), a

	;; All the remaining table entries should be filled in by the
	;; loop below, but might not be if the value in C isn't
	;; correct
	ld	a, low(ReduceModN_Broken)
	inc	e
SetupMod_InitTableLoop:
	ld	(de), a
	inc	e
	jr	nz, SetupMod_InitTableLoop

	ld	b, 8
	ld	ix, setupModTable
SetupMod_Loop:
	;; (IX) = address of buffer for shifted value
	;; (IX + 2) = corresponding LSB jump address to be stored in
	;; the modJumpTable
	ld	e, (ix + 0)
	ld	d, (ix + 1)
	push	bc
	 push	hl
	  ld	b, (hl)		; B = # bytes of modulus
	  inc	e
	  ;; Shift each byte of the modulus left appropriately
	  inc	hl
	  push	hl
	   ld	l, (hl)
	   ld	h, c
SetupMod_ShiftByte0Loop:
	   add	hl, hl
	   jr	nc, SetupMod_ShiftByte0Loop
	   ld	a, l
	   or	a
	   jr	z, SetupMod_ShiftBytesBegin1
	   jr	SetupMod_ShiftBytesBegin2

SetupMod_ShiftLoop:
	  inc	hl
	  push	hl
	   ld	l, (hl)
	   ld	h, c
SetupMod_ShiftByteLoop:
	   add	hl, hl
	   jr	nc, SetupMod_ShiftByteLoop
	   ld	a, (de)
	   or	l
SetupMod_ShiftBytesBegin2:
	   ld	(de), a
	   inc	e
SetupMod_ShiftBytesBegin1:
	   ld	a, h
	   ld	(de), a
	   pop	hl
	  djnz	SetupMod_ShiftLoop

	  ;; Store a zero at the end of the shifted buffer (for the
	  ;; benefit of the "-1" case)
	  inc	e
	  xor	a
	  ld	(de), a

	  ;; Find the actual MSB of the shifted buffer (it might be at
	  ;; DE-1 or DE-2 depending on the value and shift we're
	  ;; using)
SetupMod_FindMSBLoop:
	  dec	e
	  ld	a, (de)
	  or	a
	  jr	z, SetupMod_FindMSBLoop

	  ;; For all bytes strictly greater than A, set the
	  ;; corresponding jump address in the modJumpTable.  (That is
	  ;; to say, if MSB of a number to be reduced is greater than
	  ;; A, we know that we can safely subtract out the shifted
	  ;; modulus value we just computed.)
	  inc	a		; Assuming the high 8 bits of the
				; modulus are not all ones!
	  ld	l, a
	  ld	h, high(modJumpTable)
	  ld	a, (ix + 2)
SetupMod_FillTableLoop:
	  ld	(hl), a
	  inc	l
	  jr	nz, SetupMod_FillTableLoop

	  ;; Find the total length, in bytes, of the shifted modulus,
	  ;; and store it at the start of the buffer (TI bignum style)
	  ld	a, e
	  sub	(ix + 0)
	  ld	e, (ix + 0)
	  ld	(de), a
	  pop	hl
	 ld	de, 3
	 add	ix, de
	 pop	bc
	rrc	c
	djnz	SetupMod_Loop
	ret
