;;; -*- TI-Asm -*-
;;;
;;; DuckSign - sign Flash applications on the TI-83 Plus
;;;
;;; Copyright (C) 2010 Benjamin Moody
;;;
;;; This program is free software: you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;;;;;;;;;;;;;;; DID YOU REMEMBER TO DISABLE INTERRUPTS? ;;;;;;;;;;;;;;;;

;;; Important notes on the multiplication routines:
;;;
;;; Arguments to the multiplication routines are generally marked as "no
;;; carries".  This means that there can be no carries between the low
;;; and high byte at any point within the buffer; i.e., the MSB of the
;;; address at the end of the buffer is the same as that at the
;;; beginning of the buffer.
;;;
;;; Unless otherwise noted, input and output buffers must not overlap.
;;;
;;; Each of the larger multiplication routines needs its own scratch
;;; space (for instance, Mul64by64 uses mul_temp_64.)  The size of this
;;; area is equal to the size of the routine's output (so mul_temp_64
;;; must be 16 bytes); the area must have no internal carries, and must
;;; not overlap the other mul_temp areas.

;; FillSqrTab:
;;
;; Initialize the squaring table, used for fast multiplication.
;; Squaring table is 512 bytes and must be stored on a 256-byte
;; boundary (e.g., appData.)
;;
;; Destroys:
;; - AF, BC, DE, HL

FillSqrTab:
	ld	bc, sqrtab
	ld	h, c
	ld	l, c
	ld	de, 1
FillSqrTabLoop:
	ld	a, l
	ld	(bc), a
	inc	b
	ld	a, h
	ld	(bc), a
	dec	b
	add	hl, de
	inc	de
	inc	de
	inc	c
	jr	nz, FillSqrTabLoop
	ret

;; AtimesA2:
;;
;; Multiply unsigned 8-bit integers.
;;
;; T-states: 191.5 average (166.5 for AtimesC)
;;
;; Input:
;; - A = first multiplicand
;; - A' = second multiplicand
;; - Square table initialized by FillSqrTab
;;
;; Output:
;; - Carry flag clear
;; - HL = product
;; - A' = original value of A
;;
;; Destroys:
;; - AF, BC, DE, F'

AtimesA2:
	ld	c, a
	ex	af, af'
AtimesC:
	ld	b, a
	sub	c
	jr	nc, AtimesC_NC
	ld	a, c
	sub	b
AtimesC_NC:
	ld	l, a
	ld	h, high(sqrtab)
	ld	e, (hl)
	inc	h
	ld	d, (hl)		; DE = (x-y)^2 = x^2 + y^2 - 2xy
	ld	l, b
	ld	b, (hl)
	dec	h
	ld	a, c
	ld	c, (hl)		; BC = x^2
	ld	l, a
	ld	a, (hl)
	inc	h
	ld	h, (hl)		; HL = y^2
	ld	l, a
	sbc	hl, de		; HL = 2xy - x^2
	sbc	a, a
	add	hl, bc		; HL = 2xy
	adc	a, l		; L is even, so A bit 0 is set if
				; either the 'add' or the 'sbc'
				; carried, but not both
	rra
	rr	h
	rr	l
	ret

;; HLtimesDE:
;;
;; Multiply unsigned 16-bit integers.
;;
;; T-states: 207 average + 3 x AtimesA2
;;
;; Input:
;; - HL = first multiplicand
;; - DE = second multiplicand
;; - Square table initialized by FillSqrTab
;;
;; Output:
;; - AHLC = product
;;
;; Destroys:
;; - F, B, DE, AF', BC', DE', HL'

HLtimesDE:
	ld	a, l
	ex	af, af'
	ld	a, e
	exx
	call	AtimesA2	; L * E
	push	hl
	 exx
	 ld	a, h
	 ex	af, af'
	 ld	a, d
	 exx
	 call	AtimesA2	; H * D
	 push	hl
	  exx
	  ld	a, h
	  sub	l
	  jr	nc, HLtimesDE_nc1
	  neg
	  ex	af, af'
	  ld	a, d
	  sub	e
	  jr	nc, HLtimesDE_c1nc2
	  neg
HLtimesDE_nc1nc2:
	  call	AtimesA2	; (H-L) * (D-E)
	  xor	a
	  sub	l
	  ld	l, a
	  ld	a, 0
	  sbc	a, h
	  ld	h, a
HLtimesDE_finish:
	  pop	de
	 pop	bc
	sbc	a, a
	add	hl, de
	adc	a, d
	add	hl, bc
	adc	a, 0
	ld	d, e
	ld	e, b
	add	hl, de
	ret	nc
	inc	a
	ret

HLtimesDE_nc1:
	  ex	af, af'
	  ld	a, d
	  sub	e
	  jr	nc, HLtimesDE_nc1nc2
	  neg
HLtimesDE_c1nc2:
	  call	AtimesA2	; (H-L) * (E-D)
	  jp	HLtimesDE_finish

;; Mul32by32:
;;
;; Multiply unsigned 32-bit integers.
;;
;; Input:
;; - HL = address of first multiplicand (4 bytes, no carries)
;; - BC = address of second multiplicand (4 bytes, no carries)
;; - DE = address to store result (8 bytes, no carries)
;; - Square table initialized by FillSqrTab
;;
;; Output:
;; - DE increased by 8 bytes
;;
;; Destroys:
;; - AF, BC, HL, AF', BC', DE', HL'

Mul32by32_nc1:
	  ex	de, hl
	  pop	hl
	 ld	c, (hl)
	 inc	l
	 FAIL_IF_Z
	 ld	b, (hl)
	 inc	l
	 FAIL_IF_Z
	 ld	a, (hl)
	 inc	l
	 FAIL_IF_Z
	 ld	h, (hl)
	 ld	l, a
	 xor	a
	 sbc	hl, bc
	 jr	nc, Mul32by32_nc1nc2
	 sub	l
	 ld	l, a
	 sbc	a, a
	 sub	h
	 ld	h, a
Mul32by32_c1nc2:
	 call	HLtimesDE
	 ld	b, a
	 xor	a
	 jp	Mul32by32_finish

Mul32by32:
	push	de		; save output ptr
	 push	hl
	  push	bc
	   ;; Multiply low halves
; >> Mul16by16
	   push	de
	    ld	a, (bc)
	    ld	e, a
	    inc	c
	    FAIL_IF_Z
	    ld	a, (bc)
	    ld	d, a
	    ld	a, (hl)
	    inc	l
	    FAIL_IF_Z
	    ld	h, (hl)
	    ld	l, a
	    call HLtimesDE
	    ex	de, hl
	    pop	hl
	   ld	(hl), c
	   inc	l
	   FAIL_IF_Z
	   ld	(hl), e
	   inc	l
	   FAIL_IF_Z
	   ld	(hl), d
	   inc	l
	   FAIL_IF_Z
	   ld	(hl), a
	   inc	l
	   FAIL_IF_Z
	   ex	de, hl
; <<
	   pop	bc
	  pop	hl
	 push	hl
	  push	bc
	   inc	l
	   FAIL_IF_Z
	   inc	l
	   FAIL_IF_Z
	   inc	c
	   FAIL_IF_Z
	   inc	c
	   FAIL_IF_Z
	   ;; Multiply high halves
; >> Mul16by16
	   push	de
	    ld	a, (bc)
	    ld	e, a
	    inc	c
	    FAIL_IF_Z
	    ld	a, (bc)
	    ld	d, a
	    ld	a, (hl)
	    inc	l
	    FAIL_IF_Z
	    ld	h, (hl)
	    ld	l, a
	    call HLtimesDE
	    ex	de, hl
	    pop	hl
	   ld	(hl), c
	   inc	l
	   FAIL_IF_Z
	   ld	(hl), e
	   inc	l
	   FAIL_IF_Z
	   ld	(hl), d
	   inc	l
	   FAIL_IF_Z
	   ld	(hl), a
; <<
	   pop	hl
	  ld	c, (hl)
	  inc	l
	  FAIL_IF_Z
	  ld	b, (hl)
	  inc	l
	  FAIL_IF_Z
	  ld	a, (hl)
	  inc	l
	  FAIL_IF_Z
	  ld	h, (hl)
	  ld	l, a
	  xor	a
	  sbc	hl, bc		; HL = highX - lowX
	  jr	nc, Mul32by32_nc1
	  sub	l
	  ld	e, a
	  sbc	a, a
	  sub	h
	  ld	d, a		; DE = |highX - lowX|
	  pop	hl
	 ld	c, (hl)
	 inc	l
	 FAIL_IF_Z
	 ld	b, (hl)
	 inc	l
	 FAIL_IF_Z
	 ld	a, (hl)
	 inc	l
	 FAIL_IF_Z
	 ld	h, (hl)
	 ld	l, a
	 xor	a
	 sbc	hl, bc		; HL = highY - lowY
	 jr	nc, Mul32by32_c1nc2
	 sub	l
	 ld	l, a
	 sbc	a, a
	 sub	h
	 ld	h, a
Mul32by32_nc1nc2:
	 call	HLtimesDE
	 ex	de, hl
	 ld	b, a
	 xor	a
	 ld	h, a
	 ld	l, a
	 sub	c
	 ld	c, a
	 ld	a, h
	 sbc	hl, de
	 sbc	a, b
	 ld	b, a
	 sbc	a, a
Mul32by32_finish:
	 pop	de
	exx
	ld	b, a
	exx

	;; temp product in B':B:H:L:C
	;; add half products to temp
	ld	a, (de)
	inc	e
	FAIL_IF_Z
	add	a, c
	ld	c, a
	ld	a, (de)
	inc	e
	FAIL_IF_Z
	adc	a, l
	ld	l, a
	ld	a, (de)
	inc	e
	FAIL_IF_Z
	adc	a, h
	ld	h, a
	ld	a, (de)
	inc	e
	FAIL_IF_Z
	adc	a, b
	ld	b, a
	exx
	ld	a, 0
	adc	a, b
	ld	b, a
	exx

	ld	a, (de)
	inc	e
	FAIL_IF_Z
	add	a, c
	ld	c, a
	ld	a, (de)
	inc	e
	FAIL_IF_Z
	adc	a, l
	ld	l, a
	ld	a, (de)
	inc	e
	FAIL_IF_Z
	adc	a, h
	ld	h, a
	ld	a, (de)
	adc	a, b
	ld	b, a
	exx
	ld	a, 0
	adc	a, b
	ld	b, a
	exx

	ld	a, e
	sub	5
	FAIL_IF_C
	ld	e, a		; DE = original output ptr plus 2

	;; add temp back into final result
	ld	a, (de)
	add	a, c
	ld	(de), a
	inc	e
	FAIL_IF_Z
	ld	a, (de)
	adc	a, l
	ld	(de), a
	inc	e
	FAIL_IF_Z
	ld	a, (de)
	adc	a, h
	ld	(de), a
	inc	e
	FAIL_IF_Z
	ld	a, (de)
	adc	a, b
	ld	(de), a
	inc	e
	FAIL_IF_Z
	exx
	ld	a, b
	exx
	ld	h,d
	ld	l,e
	adc	a, (hl)
	ld	(hl), a

	inc	e
	FAIL_IF_Z
	inc	e
	FAIL_IF_Z
	ret	nc
	inc	l
	inc	(hl)
	ret

;; Mul64by64:
;;
;; Multiply unsigned 64-bit integers.
;;
;; Input:
;; - HL = address of first multiplicand (8 bytes, no carries)
;; - BC = address of second multiplicand (8 bytes, no carries)
;; - DE = address to store result (16 bytes, no carries)
;; - Square table initialized by FillSqrTab
;;
;; Output:
;; - DE increased by 16 bytes
;;
;; Destroys:
;; - AF, BC, HL, AF', BC', DE', HL'
;; - mul_temp_64

Mul64by64:
	push	de		; save output ptr
	 push	hl
	  push	bc
	   call	Mul32by32	; multiply low halves
	   pop	bc
	  pop	hl
	 push	hl
	  push	bc
	   ld	a, l
	   add	a, 4
	   FAIL_IF_C
	   ld	l, a
	   ld	a, c
	   add	a, 4
	   FAIL_IF_C
	   ld	c, a
	   call	Mul32by32	; multiply high halves
	   pop	bc
	  ld	de, mul_diffA_64
	  ld	a, 4
	  call	HalfDiffN
	  pop	bc
	 ex	af, af'
	 ld	a, 4
	 call	HalfDiffN
	 ld	c, a
	 ex	af, af'
	 xor	c
	 push	af
	  ld	hl, mul_diffA_64
	  ld	bc, mul_diffB_64
	  ld	de, mul_temp_64
	  call	Mul32by32	; temp = |highX-lowX| * |highY-lowY|
	  xor	a
	  ld	(de), a
	  pop	af
	 ;; negate temp if appropriate
	 jr	nz, Mul64by64_NoNeg
	 ld	b, 9
	 ld	hl, mul_temp_64
	 call	NegN
Mul64by64_NoNeg:
	 pop	hl
	;; HL = original output ptr
	;; add half products to temp
	ld	de, mul_temp_64
	ld	bc, 0801h	; 8 bytes half-product -> 9 bytes temp
	call	AddNtoM
	ld	de, mul_temp_64
	ld	bc, 0801h
	call	AddNtoM
	;; add temp back into final result
	ld	a, l
	sub	12
	FAIL_IF_C
	ld	l, a
	ex	de, hl		; DE = original output ptr plus 4
	ld	hl, mul_temp_64
	ld	bc, 0903h	; 9 bytes temp -> 12 bytes result
	call	AddNtoM
	inc	e
	FAIL_IF_Z
	inc	e
	FAIL_IF_Z
	inc	e
	FAIL_IF_Z
	ret

;; Mul128by128:
;;
;; Multiply unsigned 128-bit integers.
;;
;; Input:
;; - HL = address of first multiplicand (16 bytes, no carries)
;; - BC = address of second multiplicand (16 bytes, no carries)
;; - DE = address to store result (32 bytes, no carries)
;; - Square table initialized by FillSqrTab
;;
;; Output:
;; - DE increased by 32 bytes
;;
;; Destroys:
;; - AF, BC, HL, AF', BC', DE', HL'
;; - mul_temp_128

Mul128by128:
	push	de		; save output ptr
	 push	hl
	  push	bc
	   call	Mul64by64	; multiply low halves
	   pop	bc
	  pop	hl
	 push	hl
	  push	bc
	   ld	a, l
	   add	a, 8
	   FAIL_IF_C
	   ld	l, a
	   ld	a, c
	   add	a, 8
	   FAIL_IF_C
	   ld	c, a
	   call	Mul64by64	; multiply high halves
	   pop	bc
	  ld	de, mul_diffA_128
	  ld	a, 8
	  call	HalfDiffN
	  pop	bc
	 ex	af, af'
	 ld	a, 8
	 call	HalfDiffN
	 ld	c, a
	 ex	af, af'
	 xor	c
	 push	af
	  ld	hl, mul_diffA_128
	  ld	bc, mul_diffB_128
	  ld	de, mul_temp_128
	  call	Mul64by64	; temp = |highX-lowX| * |highY-lowY|
	  xor	a
	  ld	(de), a
	  pop	af
	 ;; negate temp if appropriate
	 jr	nz, Mul128by128_NoNeg
	 ld	b, 17
	 ld	hl, mul_temp_128
	 call	NegN
Mul128by128_NoNeg:
	 pop	hl
	;; HL = original output ptr
	;; add half products to temp
	ld	de, mul_temp_128
	ld	bc, 1001h	; 16 bytes half-product -> 17 bytes temp
	call	AddNtoM
	ld	de, mul_temp_128
	ld	bc, 1001h
	call	AddNtoM
	;; add temp back into final result
	ld	a, l
	sub	24
	FAIL_IF_C
	ld	l, a
	ex	de, hl		; DE = original output ptr plus 8
	ld	hl, mul_temp_128
	ld	bc, 1107h	; 17 bytes temp -> 24 bytes result
	call	AddNtoM
	ld	a, e
	add	a, 7
	FAIL_IF_C
	ld	e, a
	ret

;; Mul256by256:
;;
;; Multiply unsigned 256-bit integers.
;;
;; Input:
;; - HL = address of first multiplicand (32 bytes, no carries)
;; - BC = address of second multiplicand (32 bytes, no carries)
;; - DE = address to store result (64 bytes, no carries)
;; - Square table initialized by FillSqrTab
;;
;; Output:
;; - DE increased by 64 bytes
;;
;; Destroys:
;; - AF, BC, HL, AF', BC', DE', HL'
;; - mul_temp_256

Mul256by256:
	push	de		; save output ptr
	 push	hl
	  push	bc
	   call	Mul128by128	; multiply low halves
	   pop	bc
	  pop	hl
	 push	hl
	  push	bc
	   ld	a, l
	   add	a, 16
	   FAIL_IF_C
	   ld	l, a
	   ld	a, c
	   add	a, 16
	   FAIL_IF_C
	   ld	c, a
	   call	Mul128by128	; multiply high halves
	   pop	bc
	  ld	de, mul_diffA_256
	  ld	a, 16
	  call	HalfDiffN
	  pop	bc
	 ex	af, af'
	 ld	a, 16
	 call	HalfDiffN
	 ld	c, a
	 ex	af, af'
	 xor	c
	 push	af
	  ld	hl, mul_diffA_256
	  ld	bc, mul_diffB_256
	  ld	de, mul_temp_256
	  call	Mul128by128	; temp = |highX-lowX| * |highY-lowY|
	  xor	a
	  ld	(de), a
	  pop	af
	 ;; negate temp if appropriate
	 jr	nz, Mul256by256_NoNeg
	 ld	b, 33
	 ld	hl, mul_temp_256
	 call	NegN
Mul256by256_NoNeg:
	 pop	hl
	;; HL = original output ptr
	;; add half products to temp
	ld	de, mul_temp_256
	ld	bc, 2001h	; 32 bytes half-product -> 33 bytes temp
	call	AddNtoM
	ld	de, mul_temp_256
	ld	bc, 2001h
	call	AddNtoM
	;; add temp back into final result
	ld	a, l
	sub	48
	FAIL_IF_C
	ld	l, a
	ex	de, hl		; HL = original output ptr plus 16
	ld	hl, mul_temp_256
	ld	bc, 210Fh	; 33 bytes temp -> 48 bytes result
	call	AddNtoM
	ld	a, e
	add	a, 15
	FAIL_IF_C
	ld	e, a
	ret

;; Mul264by264:
;;
;; Multiply unsigned 264-bit integers.
;;
;; Input:
;; - HL = address of first multiplicand (33 bytes, no carries)
;; - BC = address of second multiplicand (33 bytes, no carries)
;; - DE = address to store result (66 bytes, no carries)
;; - Square table initialized by FillSqrTab
;;
;; Destroys:
;; - AF, BC, DE, HL, IX, AF', BC', DE', HL'

Mul264by264:
	push	hl
	 push	bc
	  call	Mul256by256
	  pop	bc
	 ld	hl, 32
	 add	hl, bc
	 ld	a, (hl)
	 pop	ix
	push	ix
	 push	bc
	  ld	c, (ix + 32)
	  push	de
	   call	AtimesC
	   pop	de
	  ld	a, l
	  ld	(de), a
	  inc	de
	  ld	a, h
	  ld	(de), a
	  inc	de
	  pop	hl
	 push	hl
	  call	Mul256byHigh8
	  pop	ix
	 pop	hl
	call	Mul256byHigh8
	ret

Mul256byHigh8:
	;; multiply 32 bytes at HL by (IX + 32) and add result to
	;; buffer at DE
	ld	a, (ix + 32)
	or	a
	ret	z
	push	de
	 ex	af, af'
	 ex	de, hl
	 ld	bc, -34
	 add	hl, bc		; HL = output buffer + 32
	 ld	b, 32
Mul256byHigh8_Loop:
	 ;; inputs A', (DE) output (HL)
	 ld	a, (de)
	 ex	af, af'
	 exx
	 call	AtimesA2
	 ld	a, l
	 exx
	 add	a, (hl)
	 ld	(hl), a
	 inc	hl
	 exx
	 ld	a, h
	 exx
	 adc	a, (hl)
	 ld	(hl), a
	 call	c, CarryInc
	 inc	de
	 djnz	Mul256byHigh8_Loop
	 pop	de
	ret

;; HalfDiffN:
;;
;; Compute absolute value of the difference between two integers
;; stored in adjacent memory buffers.
;;
;; Input:
;; - A = length of each input, bytes
;; - BC = address of inputs (2*A bytes, no carries)
;; - DE = address of buffer to store result (A bytes, no carries)
;;
;; Output:
;; - Result stored at DE
;; - DE increased by A bytes
;; - A = 0 if first input is less than the second, A = input A otherwise
;;
;; Destroys:
;; - AF, BC, HL, DE

HalfDiffN:
	ld	h, 0
	ld	l, a
	add	hl, bc
	push	de
	 exx
	 ld	b, a
	 ld	c, a
HalfDiffN_loop:
	 exx
	 ld	a, (bc)
	 sbc	a, (hl)
	 ld	(de), a
	 inc	l
	 FAIL_IF_Z
	 inc	c
	 FAIL_IF_Z
	 inc	e
	 FAIL_IF_Z
	 exx
	 djnz	HalfDiffN_loop
	 ld	a, c
	 exx
	 pop	hl
	ret	nc
	ld	b, a
	;; fall through

;; NegN:
;;
;; Negate an integer stored at address HL.
;;
;; Input:
;; - B = length of number, bytes
;; - HL = address of input (B bytes, no carries)
;;
;; Output:
;; - A = 0
;;
;; Destroys:
;; - AF, BC, HL

NegN:
	xor	a
	ld	c, a
NegN_Loop:
	sbc	a, (hl)
	ld	(hl), a
	inc	l
	FAIL_IF_Z
	ld	a, c
	djnz	NegN_Loop
	ret

