;;; -*- TI-Asm -*-
;;;
;;; DuckSign - sign Flash applications on the TI-83 Plus
;;;
;;; Copyright (C) 2010 Benjamin Moody
;;;
;;; This program is free software: you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;; ValidateSignature:
;;
;; Check that a Rabin signature is valid for a given MD5 hash.
;;
;; Input:
;; - Either interrupts must be disabled, or a protective IM 2 handler
;;   installed
;; - (MD5Hash) = hash of application (16 bytes)
;; - Carry flag set if ON key was pressed (calculation aborted)
;; - (rabinSignature_length) = Signature (TI bignum format)
;; - (rabinF_length) = "f" value (TI bignum format)
;;
;; Output:
;; - Zero flag set if signature is correct
;;
;; Destroys:
;; - AF, BC, DE, HL, IX, AF', BC', DE', HL'
;; - appData
;; - ramCode
;; - MD5Buffer
;; - OP1-OP6
;; - various scratch buffers in saveSScreen and tempSwapArea

ValidateSignature:
	xor	a
	BCALL	_SetupAppPubKey	; defaults to key 0104

	BCALL	_TransformHash	; stores "transformed" value as bignum
				; in MD5Buffer

  ifndef INTERNAL_VALIDATE

	ld	hl, rabinSignature_length
	BCALL	_RabinValidate	; stores square of signature as bignum
				; in OP1-OP6

	ld	hl, OP1
	ld	de, MD5Buffer
	BCALL	_BigNumCompare
	ret

  else

	call	FillSqrTab
	call	SetupMod_P0104
	call	ValidateSignatureMod
	ret	nz
	call	SetupMod_Q0104
ValidateSignatureMod:
	ld	hl, rabinSignature_length
	ld	de, rabinTemp1
	call	CopyBignumPad64
	LOGNUM	"s = ", rabinTemp1 + 1, 64

	push	hl
	 ld	a, 64
	 call	ReduceModN
	 pop	hl
	ld	de, rabinTemp2
	push	de
	 ld	b, h
	 ld	c, l
	 call	Mul264by264
	 pop	hl
	push	hl
	 ld	a, 66
	 call	ReduceModN
	 pop	hl
	call	ReduceModStrict	; (rabinTemp2) = squared signature
	push	hl

	 ld	hl, MD5Buffer
	 ld	de, rabinTemp1
	 call	CopyBignumPad64
	 LOGNUM	"t = ", rabinTemp1 + 1, 64

	 push	hl
	  ld	a, 64
	  call	ReduceModN
	  pop	hl
	 call	ReduceModStrict	; (rabinTemp1 + 1) = transformed hash
	 pop	de

	LOGNUM	"r1 = ", rabinTemp1 + 1, 33
	LOGNUM	"r2 = ", rabinTemp2, 33
	ld	b, 33
	jp	CompStrsN

  endif

;; GenerateSignature:
;;
;; Compute Rabin signature for a given MD5 hash.
;;
;; Input:
;; - Either interrupts must be disabled, or a protective IM 2 handler
;;   installed
;; - (MD5Hash) = hash of application (16 bytes)
;; - (penRow) = Y position to display progress bar (set to -1 to
;;   disable progress bar)
;;
;; Output:
;; - Carry flag set if ON key was pressed (calculation aborted)
;; - (rabinSignature_length) = Signature (TI bignum format)
;; - (rabinF_length) = "f" value (TI bignum format)
;;
;; Destroys:
;; - AF, BC, DE, HL, IX, AF', BC', DE', HL'
;; - appData
;; - ramCode
;; - various scratch buffers in saveSScreen and tempSwapArea

GenerateSignature:
	LOGNUM	"Hash = ", MD5Hash, 16

	ld	hl, 778
	call	InitProgress
	call	FillSqrTab

	call	SetupMod_Q0104

	;; Set initial m value
	ld	hl, rabinM
	ld	(hl), 1
	inc	hl
	ex	de, hl
	ld	hl, MD5Hash
	call	Mov16B
	ex	de, hl
	ld	b, 16
	BCALL	_ClrLp

	;; Compute Y = M ^ ((Q+1) / 4) mod Q
	ld	hl, rabinM
	ld	de, rabinY
	call	RabinSqrt
	ret	c

	LOGNUM	"M = ", rabinM, 33
	LOGNUM	"Y = ", rabinY, 33

	;; Compute Y * Y mod Q and check if this equals M
	ld	hl, rabinY
	ld	de, rabinTemp1
	call	RabinSquare
	call	ReduceModStrict

	LOGNUM	"Y^2 = ", rabinTemp1, 32

	ld	de, rabinM
	ld	b, 33
	call	CompStrsN
	ld	a, 2		; f = 2 means a factor of 1
	jr	z, GenerateSignature_Y_OK

	;; M is a nonresidue mod Q -> multiply M by -1
	;; (note that Y^2 is already congruent to -M mod Q)
	ld	bc, rabinP0104
	ld	hl, rabinM
	ld	d, h
	ld	e, l
	ld	a, 32
	call	SubN		; M = P - M (M is always less than P)

	ld	a, 1		; f = 1 means a factor of -1
GenerateSignature_Y_OK:
	ld	(rabinF), a

	call	SetupMod_P0104

	;; Compute X = M ^ ((P+1) / 4) mod P
	ld	hl, rabinM
	ld	de, rabinX
	call	RabinSqrt
	ret	c

	LOGNUM	"M' = ", rabinM, 33
	LOGNUM	"X  = ", rabinX, 33

	;; Compute X * X mod P and check if this equals M
	ld	hl, rabinX
	ld	de, rabinTemp1
	call	RabinSquare
	call	ReduceModStrict

	LOGNUM	"X^2 = ", rabinTemp1, 32

	ld	de, rabinM
	ld	b, 32
	call	CompStrsN
	jr	z, GenerateSignature_X_OK

	;; M is a nonresidue mod P -> implicitly multiply M by 2
	;; (by multiplying X and Y by pseudo-square-roots of 2)

	ld	hl, rabinF
	ld	a, (hl)
	xor	1
	ld	(hl), a		; f = 0 (1 xor 1) means a factor of -2
				; f = 3 (2 xor 1) means a factor of 2

	;; X = X * alpha mod P
	ld	hl, rabinX
	ld	bc, rabinAlpha0104
	ld	de, rabinTemp1
	call	Mul256by256Mod
	ld	de, rabinX
	call	Mov33B

	call	SetupMod_Q0104
	;; Y = Y * beta mod Q
	ld	hl, rabinY
	ld	bc, rabinBeta0104
	ld	de, rabinTemp1
	call	Mul264by264Mod
	ld	de, rabinY
	call	Mov33B
	jr	GenerateSignature_Finish

GenerateSignature_X_OK:
	call	SetupMod_Q0104
	call	UpdateProgress
	call	UpdateProgress	; we skipped two multiplications in
				; this case
GenerateSignature_Finish:

	LOGNUM	"X' = ", rabinX, 32
	LOGNUM	"Y'' = ", rabinY, 33

	;; Compute Z = (Y - X) * pi mod Q
	xor	a
	ld	(rabinX + 32), a
	ld	bc, rabinY
	ld	hl, rabinX
	ld	de, rabinTemp1
	call	SubtractMod

	LOGNUM	"Y - X mod Q = ", rabinTemp1, 33

	ld	hl, rabinTemp1
	ld	bc, rabinPi0104
	ld	de, rabinTemp2
	call	Mul264by264Mod
	call	ReduceModStrict

	LOGNUM	"Z = ", rabinTemp2, 33

	;; Compute S = Z * P + X (no modulus)
	ld	bc, rabinP0104
	ld	de, rabinSignature
	call	Mul264by264	; note that Z may be greater than
				; 2^256, but the result of this
				; multiplication is less than 2^512.

	LOGNUM	"Z * P = ", rabinSignature, 66

	ld	hl, rabinX
	ld	de, rabinSignature
	ld	bc, 2020h
	call	AddNtoM

	ld	hl, rabinSignature_length
	ld	(hl), 64
	call	ShortenBignum

	ld	hl, rabinF_length
	ld	(hl), 1
	call	ShortenBignum

	LOGNUM	"Signature = ", rabinSignature, 64
	LOGNUM	"f = ", rabinF, 1

	jp	UpdateProgress

;; RabinSqrt:
;;
;; Compute the pseudo-square-root of an integer in the current
;; modulus.  If p is a prime congruent to 3 mod 4, the
;; pseudo-square-root of x, mod p, is defined as x raised to the
;; (p+1)/4-th power.
;;
;; (If x is a quadratic residue, the square of the pseudo-square-root,
;; mod p, is x.  Otherwise the square of the pseudo-square-root is
;; -x.)
;;
;; Input:
;; - HL = input value to take pseudo-square-root of (33 bytes)
;; - DE = buffer to store result (33 bytes, no carries)
;; - Modulus values set up using SetupMod; modulus must be a prime, 32
;;   or 33 bytes long, and congruent to 3 mod 4
;;
;; Output:
;; - Value stored in output buffer, and is 33 bytes and zero padded
;;
;; Destroys:
;; - AF, BC, DE, IX, AF', BC', DE', HL'
;; - rabinTemp1
;; - rabinTemp2
;; - rabinSqrtOutput

RabinSqrt:
	;; initialize both output and rabinTemp1 to input
	push	de
	 call	Mov33B
	 pop	hl
	ld	(rabinSqrtOutput), hl
	ld	de, rabinTemp1
	call	Mov33B

	ld	a, (modulusLength)
	ld	b, a
	ld	ix, (modulus)
	ld	c, 4		; start at bit 2 of the prime
	ld	hl, rabinTemp1
	ld	de, rabinTemp2
	jr	RabinSqrt_Begin

RabinSqrt_Loop:
	;; square contents of one temp buffer, store in the other temp buffer
	push	bc
	 push	hl
	  call	RabinSquare
	  pop	de
	 pop	bc
	ret	c
RabinSqrt_Begin:
	;; HL = current temp buffer
	;; DE = alternate buffer
	;; IX = position in exponent
	;; B = number of bytes left
	;; C = current bit mask
	ld	a, (ix + 0)
	and	c
	jr	z, RabinSqrt_ZeroBit

	;; multiply contents of temp buffer by current output value;
	;; store result temporarily in the alternate buffer, then copy
	;; back to the output buffer (we could potentially speed this
	;; up by rotating between 3 buffers instead, but copying isn't
	;; very significant compared to the other stuff we're doing)
	push	bc
	 push	hl
	  ld	bc, (rabinSqrtOutput)
	  call	RabinMultiply
	  LOGNR	"... ", h, l, 0, 33
	  push	hl
	   ld	de, (rabinSqrtOutput)
	   call	nc, Mov33B
	   pop	de
	  pop	hl
	 pop	bc
	ret	c

	;; stop if we've reached the end of the exponent (B = 1 and no
	;; further bits set in this byte)
	ld	a, c
	add	a, a
	dec	a
	cpl
	and	(ix + 0)
	dec	b
	or	b
	ret	z
	inc	b
RabinSqrt_ZeroBit:
	rlc	c
	jr	nc, RabinSqrt_Loop
	inc	ix
	djnz	RabinSqrt_Loop
	xor	a
	ret

RabinSquare:
	ld	b, h
	ld	c, l
RabinMultiply:
	ld	a, (modulusLength)
	rra
	jr	nc, Mul256by256Mod
	;; fall through

;; Mul264by264Mod:
;;
;; Multiply two numbers, and subtract out (most) multiples of the
;; current modulus.
;;
;; Input:
;; - HL = first multiplicand (33 bytes, no carries)
;; - BC = second multiplicand (33 bytes, no carries)
;; - DE = output buffer (66 bytes, no carries)
;; - Modulus values set up using SetupMod
;;
;; Output:
;; - Carry flag set if ON pressed
;; - Value stored in output buffer, and is 32 bytes and zero padded
;; - HL = address of output buffer
;;
;; Destroys:
;; - AF, BC, DE, AF', BC', DE', HL'

Mul264by264Mod:
	push	de
	 push	ix
	  call	Mul264by264
	  ld	a, 66
	  jr	MulNbyNMod_Finish

;; Mul256by256Mod:
;;
;; Multiply two numbers, and subtract out (most) multiples of the
;; current modulus.
;;
;; Input:
;; - HL = first multiplicand (32 bytes, no carries)
;; - BC = second multiplicand (32 bytes, no carries)
;; - DE = output buffer (64 bytes, no carries)
;; - Modulus values set up using SetupMod
;;
;; Output:
;; - Carry flag set if ON pressed
;; - Value stored in output buffer, and is 32 bytes and zero padded
;; - HL = address of output buffer
;;
;; Destroys:
;; - AF, BC, DE, AF', BC', DE', HL'

Mul256by256Mod:
	push	de
	 push	ix
	  call	Mul256by256
	  ld	a, 64
MulNbyNMod_Finish:
	  pop	ix
	 pop	hl
	call	CheckOnAbort
	ret	c
	push	hl
	 call	ReduceModN
	 call	UpdateProgress
	 pop	hl
	jp	CheckOnAbort

