;;; -*- TI-Asm -*-
;;;
;;; DuckSign - sign Flash applications on the TI-83 Plus
;;;
;;; Copyright (C) 2010 Benjamin Moody
;;;
;;; This program is free software: you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;; GetCurAppSignaturePtr:
;;
;; Get a pointer to the signature, or the place where the signature
;; should be.
;;
;; Input:
;; - (appStartPage) = starting page
;; - (appNumPages) = number of pages
;; - (appFinalPageLength) = length of final page
;;
;; Output:
;; - A = port 6 = last page of app
;; - HL = address of signature block
;;
;; Destroys:
;; - F

GetCurAppSignaturePtr:
	ld	hl, (appStartPage)
	ld	a, l
	sub	h
	inc	a
	out	(6), a
	ld	hl, (appFinalPageLength)
	set	6, h
	ret

;; ReadAppSignature:
;;
;; Retrieve signature of the current app.
;;
;; Input:
;; - (appStartPage) = starting page
;; - (appNumPages) = number of pages
;; - (appFinalPageLength) = length of final page
;;
;; Output:
;; - Zero flag set if signature is empty (app is not currently valid
;;   but can be signed in place)
;; - Carry flag set if garbage data present (app is not currently
;;   valid and cannot be signed in place: note that it's probably a
;;   bad idea to attempt to sign the app in this case)
;; - Carry and zero flags clear if app already has a signature (may or
;;   may not be currently valid, but cannot be signed in place)
;; - If app has a signature:
;;   - (rabinSignature_length) = current signature (TI bignum format)
;;   - (rabinF_length) = current f value (TI bignum format)
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - Port 6

ReadAppSignature:
	call	GetCurAppSignaturePtr
	push	hl
	 call	CheckSignatureEmpty
	 pop	hl
	ret	z

	ld	a, (hl)
	cp	2
	scf
	ret	nz
	inc	hl
	ld	a, (hl)
	or	20h		; This is not correct, but we'll
				; permit it for now because current
				; versions of Axe use a value of 0Dh
				; here.  Don't rely on this behavior;
				; it will probably change in the
				; future.
	cp	2Dh
	scf
	ret	nz
	inc	hl

	ld	a, 64
	cp	(hl)
	ret	c
	ld	de, rabinSignature_length
	call	CopyBignum

	ld	a, 1
	ld	de, rabinF_length
	call	CopyBignumPad
	or	h		; Again, allowing invalid values
				; because current versions of Axe put
				; nonsense values there.  Don't do
				; this.
	ret

;; CheckSignatureEmpty:
;;
;; Check if signature block is empty or a skeleton (i.e., we can write
;; a new signature in place.)
;;
;; Input:
;; - HL = address of (potential) signature block
;;
;; Output:
;; - Zero flag set if block is "empty"
;;
;; Destroys:
;; - AF, HL, B

CheckSignatureEmpty:
	ld	a, (hl)
	cpl
	and	2
	ret	nz
	inc	hl
	ld	a, (hl)
	cpl
	and	2Dh
	ret	nz
	inc	hl
	ld	a, (hl)
	cpl
	and	40h
	ret	nz
	inc	hl
	ld	b, 64
CheckSignatureEmpty_Loop:
	ld	a, (hl)
	inc	a
	ret	nz
	inc	hl
	djnz	CheckSignatureEmpty_Loop
	ld	a, (hl)
	cpl
	and	1
	ret	nz
	inc	hl
	ld	a, (hl)
	cpl
	and	3
	ret

;; BuildAppSignature:
;;
;; Write application signature block to a RAM buffer.
;;
;; Input:
;; - DE = address of buffer to store signature block (69 bytes)
;; - (rabinSignature_length) = current signature (TI bignum format)
;; - (rabinF_length) = current f value (TI bignum format)
;;
;; Output:
;; - HL = input DE
;; - B = length of signature block
;;
;; Destroys:
;; - AF, C, DE

BuildAppSignature:
	push	de
	 ld	a, 2
	 ld	(de), a
	 inc	de
	 ld	a, 2Dh
	 ld	(de), a
	 inc	de
	 ld	hl, rabinSignature_length
	 call	CopyBignum
	 ld	hl, rabinF_length
	 call	CopyBignum
	 pop	hl
	ld	a, e
	sub	l
	ld	b, a
	ret

;; EraseAppSignature:
;;
;; Erase app signature if necessary.
;;
;; Input:
;; - (appStartPage) = starting page (must be in Flash)
;; - (appNumPages) = number of pages
;; - (appFinalPageLength) = length of final page
;;
;; Output:
;; - Zero flag clear if erasing fails
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - Port 6
;; - appData
;; - ramCode
;; - Interrupts disabled

EraseAppSignature:
	call	GetCurAppSignaturePtr
	call	CheckSignatureEmpty
	ret	z

	call	TestExtraRAM
	ld	d, 84h
	jr	nc, EraseAppSignatureWithTemp
	BCALL	_FindSwapSector
	ld	d, a
	out	(6), a
	ld	a, (4000h)
	ld	e, a
	push	de
	 call	EraseAppSignatureWithTemp
	 pop	de
	;; even if something goes horribly wrong in one or both copy
	;; operations, we must be sure to erase the temp sector when
	;; we're done
	push	af
	 ld	a, d
	 call	EraseUserFlashPage
	 jr	nz, EraseAppSignature_LastEraseFailed
	 ld	a, d
	 ld	b, e
	 ld	de, 4000h
	 BCALL	_WriteAByteSafe
EraseAppSignature_LastEraseFailed:
	 pop	af
	ret

EraseAppSignatureWithTemp:
	call	GetCurAppSignaturePtr
	and	~3
	ld	e, a
	push	de
	 call	CopyPartialSector
	 pop	bc
	ret	nz
	ld	e, b
	ld	d, c
	;; fall through

;; CopyPartialSector:
;;
;; Copy 4 pages of memory.  If the destination pages are in Flash,
;; erase them first.  If one of the destination pages is the final
;; page of the current app, don't copy the existing signature (if
;; any.)
;;
;; Input:
;; - E = starting page of source sector
;; - D = starting page of destination sector (if in Flash, must be a
;;   multiple of 4)
;; - (appStartPage) = current app's starting page
;; - (appNumPages) = current app's number of pages
;; - (appFinalPageLength) = length of final page
;;
;; Output:
;; - Zero flag clear if something went wrong
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - Port 6
;; - appData
;; - ramCode
;; - Interrupts disabled

CopyPartialSector:
	call	GetFlashPageMask
	cp	d
	jr	c, CopyPartialSector_NoErase
	ld	a, d
	call	EraseUserFlashPage
	ret	nz
CopyPartialSector_NoErase:
	call	GetCurAppSignaturePtr
	ld	c, a
	ld	b, 4
CopyPartialSector_Loop:
	push	bc
	 ld	a, d
	 cp	c
	 ld	hl, 4000h
	 ld	b, h
	 ld	c, l
	 jr	nz, CopyPartialSector_FullPage
	 ld	bc, (appFinalPageLength)
CopyPartialSector_FullPage:
	 inc	b
	 jr	CopyPartialSector_BlockBegin

CopyPartialSector_BlockLoop:
	 push	bc
	  ld	bc, 256
	  call	CopyPageBlock
	  pop	bc
	 jr	nz, CopyPartialSector_Error
CopyPartialSector_BlockBegin:
	 djnz CopyPartialSector_BlockLoop
	 inc	c
	 dec	c
	 call	nz, CopyPageBlock
CopyPartialSector_Error:
	 pop	bc
	ret	nz
	inc	d
	inc	e
	djnz	CopyPartialSector_Loop
	xor	a
	ret

CopyPageBlock:
	ld	a, e
	out	(6), a
	push	de
	 call	GetFlashPageMask
	 cp	d
	 ld	a, d
	 ld	de, appData
	 push	de
	  push	hl
	   push	bc
	    ldir
	    pop	bc
	   pop	de
	  ex	(sp), hl
	  jr	c, CopyPageBlock_ToRAM
	  call	WriteUserFlash
	  jr	CopyPageBlock_Done

CopyPageBlock_ToRAM:
	  out	(6), a
	  ldir
	  xor	a
CopyPageBlock_Done:
	  pop	hl
	 pop	de
	ret

;; WriteAppSignature:
;;
;; Write application signature to an existing Flash application.
;;
;; Input:
;; - (appStartPage) = starting page (must be in Flash)
;; - (appNumPages) = number of pages
;; - (appFinalPageLength) = length of final page
;; - (rabinSignature_length) = current signature (TI bignum format)
;; - (rabinF_length) = current f value (TI bignum format)
;;
;; Output:
;; - Zero flag clear if data cannot be written to Flash (either due to
;;   a hardware error, or because of conflicting data already present)
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - Port 6
;; - appData
;; - ramCode
;; - Interrupts disabled

WriteAppSignature:
	call	GetCurAppSignaturePtr
	push	hl
	 inc	hl
	 inc	hl
	 ;; If a skeleton signature is in place (with length 64), write
	 ;; our signature as 64 bytes as well; otherwise, use the
	 ;; canonical form.

	 ;; (We always use the canonical form for the f value; this
	 ;; should work fine even if a skeleton f value is present.)
	 ld	a, (hl)
	 cp	64
	 jr	nz, WriteAppSignature_UseCanonical
	 ld	(rabinSignature_length), a
WriteAppSignature_UseCanonical:
	 ld	de, appData
	 call	BuildAppSignature
	 pop	de
	;; fall through

;; CheckWriteUserFlash:
;;
;; Write data to user archive/app memory.  Verify that there are no
;; conflicts before attempting to write anything.
;;
;; Input:
;; - Port 6 = destination page (must be in user Flash; no page
;;   overflows allowed)
;; - HL = address of data to write (must be in RAM)
;; - DE = address to write (must be between 4000 and 7FFF)
;; - B = number of bytes to write
;;
;; Output:
;; - Zero flag clear if data cannot be written to Flash (either due to
;;   a hardware error or because of conflicting data already present)
;;
;; Destroys:
;; - AF, BC, DE, HL, others?
;; - Port 6
;; - ramCode
;; - Interrupts disabled

CheckWriteUserFlash:
	push	bc
	 push	hl
	  push	de
CheckWriteUserFlash_Loop:
	   ld	a, (de)
	   cpl
	   and	(hl)
	   jr	nz, CheckWriteUserFlash_Failed
	   inc	de
	   inc	hl
	   djnz	CheckWriteUserFlash_Loop
	   bit	7, d
CheckWriteUserFlash_Failed:
	   pop	de
	  pop	hl
	 pop	bc
	ret	nz
	ld	c, b
	ld	b, 0
	in	a, (6)
	;; fall through

;; WriteUserFlash:
;;
;; Write data to Flash within user app/archive space.
;;
;; Input:
;; - A = destination page
;; - DE = destination address (must be in Flash)
;; - HL = source address (must be in RAM)
;; - BC = number of bytes to write
;;
;; Output:
;; - Zero flag clear if address is not valid or data could not be
;;   written
;;
;; Destroys:
;; - F, BC, HL
;; - ramCode
;; - Interrupts disabled

WriteUserFlash:
	di
	call	CheckUserFlashPage
	ret	c
	BCALL	_WriteFlashSafe
	ret

;; EraseUserFlashPage:
;;
;; Erase a Flash sector located in user app/archive space.
;;
;; Input:
;; - A = any page number within the given sector
;;
;; Output:
;; - Zero flag clear if address is not valid or data could not be
;;   erased
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - ramCode
;; - Interrupts disabled

EraseUserFlashPage:
	di
	call	CheckUserFlashPage
	ret	c
	push	de
	 BCALL	_EraseFlashPage
	 pop	de
	ret

