;;; -*- TI-Asm -*-
;;;
;;; DuckSign - sign Flash applications on the TI-83 Plus
;;;
;;; Copyright (C) 2010 Benjamin Moody
;;;
;;; This program is free software: you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License as
;;; published by the Free Software Foundation; either version 3 of the
;;; License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;; GetFlashPageMask:
;;
;; Determine the last Flash page, or the mask of bits used for Flash
;; pages, for this calculator model.
;;
;; Output:
;; - A = last Flash page
;;
;; Destroys:
;; - F

GetFlashPageMask:
	in	a, (2)
	add	a, a
	ld	a, 1Fh
	ret	nc
	in	a, (21h)
	and	3
	ld	a, 7Fh
	ret	nz
	rra
	ret

;; SetFlashPageD:
;;
;; Switch the Flash page mapped in port 6.
;;
;; Input:
;; - D = generic page number
;;
;; Output:
;; - A = port 6 = actual page number for this model
;;
;; Destroys:
;; - F

SetFlashPageD:
	call	GetFlashPageMask
	and	d
	out	(6), a
	ret

;; CheckOnAbort:
;;
;; Check if the ON key is pressed.
;;
;; Output:
;; - Carry flag set if ON pressed
;;
;; Destroys:
;; - F

CheckOnAbort:
	push	bc
	 ld	b, a
	 in	a, (4)
	 and	8
	 sub	1
	 ld	a, b
	 pop	bc
	ret

;; CompStrsN:
;;
;; Compare two binary strings.
;;
;; Input:
;; - HL = address of first string
;; - DE = address of second string
;; - B = number of bytes to compare (at least 1)
;;
;; Output:
;; - Zero flag set if strings match
;; - Carry flag set if string at DE is lexicographically earlier
;;
;; Destroys:
;; - AF

CompStrsN:
	ld	a, (de)
	cp	(hl)
	ret	nz
	inc	hl
	inc	de
	djnz	CompStrsN
	ret

;; CarryInc:
;;
;; Increment the number at address HL + 1.
;;
;; Input:
;; - HL = address of number minus 1
;;
;; Destroys:
;; - F

CarryInc:
	push	hl
CarryInc_Loop:
	 inc	hl
	 inc	(hl)
	 jr	z, CarryInc_Loop
	 pop	hl
	ret


;; CheckBattery:
;;
;; Display an error message if batteries are low.
;;
;; Output:
;; - Carry flag set if batteries are good
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - (penRow), (penCol), (curRow), (curCol)

CheckBattery:
	BCALL	_Chk_Batt_Low
	scf
	ret	nz
	ld	hl, batteryLowMsg
	;; fall through

;; ShowMessageWait:
;;
;; Display a message, display "Press a key", and wait for user to do so.
;;
;; Input:
;; - HL = address of message to display
;;
;; Output:
;; - A = scan code
;;
;; Destroys:
;; - F, BC, DE, HL
;; - (penRow), (penCol)

ShowMessageWait:
	call	DispMessage
	ld	hl, 301Dh
	ld	(penCol),hl
	ld	hl, pressAKeyStr
	BCALL	_VPutS
	;; fall through

;; WaitCSC:
;;
;; Wait for a key to be pressed.
;;
;; Output:
;; - A = scan code
;;
;; Destroys:
;; - F, HL

WaitCSC:
	ei
	halt
	BCALL	_GetCSC
	or	a
	jr	z, WaitCSC
	ret

;; DispMessage:
;;
;; Clear screen and display a message.  The message is a
;; zero-terminated string, with a few special byte values: the byte
;; value 1 stands for a line feed, and the bytes 2 through 31 stand
;; for common substrings (listed in the messageStrs table.)
;;
;; Normally, the name of the currently selected application is also
;; displayed.  If the message string begins with a zero byte, this
;; behavior is suppressed.
;;
;; Input:
;; - HL = address of message to display
;;
;; Destroys:
;; - AF, BC, DE, HL

DispMessage:
	push	hl
	 BCALL	_ClrLCDFull
	 ld	hl, 1
	 ld	(curRow), hl
	 pop	de
	ld	a, (de)
	or	a
	jr	z, DispMessage_Loop
	dec	de
	ld	hl, appNameStr
	BCALL	_PutS
	push	de
	 ld	a, (appStartPage)
	 call	GetAppName
	 BCALL	_PutS
	 pop	de
DispMessage_NewLine:
	BCALL	_NewLine
DispMessage_Loop:
	inc	de
	ld	a, (de)
	or	a
	ret	z
	cp	' '
	jr	c, DispMessage_Special
	BCALL	_PutC
	jr	DispMessage_Loop

DispMessage_Special:
	dec	a
	jr	z, DispMessage_NewLine
	ld	b, a
	ld	hl, messageStrs
DispMessage_FindStrLoop:
	ld	c, 255
	xor	a
	cpir
	djnz	DispMessage_FindStrLoop
	BCALL	_PutS
	jr	DispMessage_Loop


;; DispProgressMessage:
;;
;; Clear screen and display a message.  Display "Press ON to cancel"
;; below it, and set the pen position for a progress bar.
;;
;; Input:
;; - HL = address of message to display
;;
;; Destroys:
;; - AF, BC, DE, HL

DispProgressMessage:
	call	DispMessage
	ld	hl, 3311h
	ld	(penCol),hl
	ld	hl, pressOnToCancelStr
	BCALL	_VPutS
	ld	a, 2Ch
	ld	(penRow), a
	ret

;; TestExtraRAM:
;;
;; Check if all RAM pages 82h to 87h are usable.
;;
;; Output:
;; - Carry flag set if extra RAM pages are not usable
;;
;; Destroys:
;; - AF, BC, DE, HL
;; - OP1
;; - Port 6
;; - Interrupts disabled

TestExtraRAM:
	in	a, (2)
	sub	80h
	ret	c
	di
	ld	hl, 4000h
	ld	bc, 0880h
	ld	de, OP1
	push	bc
TestExtraRAM_FillLoop:
	 ld	a, c
	 out	(6), a
	 ld	a, (hl)
	 ld	(de), a
	 ld	(hl), c
	 inc	c
	 inc	de
	 djnz	TestExtraRAM_FillLoop
	 pop	bc
TestExtraRAM_CheckLoop:
	ld	a, c
	out	(6), a
	jr	c, TestExtraRAM_Bad
	xor	(hl)
	add	a, -1
TestExtraRAM_Bad:
	inc	c
	djnz	TestExtraRAM_CheckLoop
	ld	b, 8
TestExtraRAM_FixLoop:
	dec	c
	dec	de
	ld	a, c
	out	(6), a
	ld	a, (de)
	ld	(hl), a
	djnz	TestExtraRAM_FixLoop
	ret

Unlock:
	di
	push	iy
	 ld	de, 7B09h
	 ld	a, e
	 ld	(arcInfo), a
	 out	(6), a
	 ld	a, (de)
	 ld	b, a
	 push	bc
	  push	de
	   call	SetFlashPageD
	   ld	a, e
	   BCALL _WriteAByteSafe
	   ld	hl, Unlock_Recover
	   ld	de, pagedBuf
	   call	Mov16B
	   ld	(pagedBuf + 2), sp
	   call	Unlock_Main
	   pop	af
	  pop	bc
	 jr Unlock_Done

Unlock_Main:
	ld	a, (_GetVariableData + 2)
	ld	d, a
	ld	hl, ramCode
	ld	bc, 304
Unlock_Loop1:
	ld	a, 0EDh
	cpir
	ret	nz
	ld	a, (hl)
	sub	0A0h
	jr	nz, Unlock_Loop1
	ld	c, (hl)
	ld	b, a
	inc	hl
	ld	(iMathPtr5), hl
	ld	hl, (_GetVariableData)
	inc	a
Unlock_Data:
	ld	(pagedCount), a
	or	a
	call	nz, SetFlashPageD
Unlock_Loop2:
	ld	a, 3Ah
	cpir
	ret	nz
	ld	de, Unlock_Data + 1
Unlock_Loop3:
	ld	a,(de)
	cp	(hl)
	jr	nz, Unlock_Loop2
	inc	hl
	inc	de
	cp	0C4h
	jr	nz, Unlock_Loop3
	BCALL	_LdHLind
	ld	iy, 56h - 25h
Unlock_Recover:
	jp	(hl)
	   ld	sp, 0
	   pop	hl
	  pop	af
	 ld	(hl), a
Unlock_WriteByte:
	 cp	(hl)
	 jr	nz, Unlock_WriteByte
	 ld	(hl), 0F0h
Unlock_Done:
	 pop	iy
	ret

;; CopyBignum:
;;
;; Copy a TI-format bignum.
;;
;; Input:
;; - HL = address of number
;; - DE = address of buffer to place result
;;
;; Output:
;; - HL = address of byte following number
;; - DE = address of byte following result
;;
;; Destroys:
;; - F, BC

CopyBignum:
	ld	c, (hl)
	ld	b, 0
	inc	bc
	jr	MovNB

;; CopyBignumPad:
;;
;; Copy a TI-format bignum, and pad the output buffer with zeroes.
;;
;; Input:
;; - HL = address of number
;; - DE = address of buffer to place result
;; - A = size of output buffer minus 1
;;
;; Output:
;; - Carry flag set if input is too large to fit in buffer
;; - HL = address of LSB of result (input DE plus 1)
;;
;; Destroys:
;; - AF, B, DE

CopyBignumPad64:
	ld	a, 64
CopyBignumPad:
	sub	(hl)
	ret	c
	push	de
	 call	CopyBignum
	 pop	hl
	inc	hl
	or	a
	ret	z
	ex	de, hl
	ld	b, a
	BCALL	_ClrLp
	ex	de, hl
	ret

;; Mov33B:
;;
;; Copy 33 bytes from one memory buffer to another.
;;
;; Input:
;; - HL = source
;; - DE = destination
;;
;; Output:
;; - HL and DE each incremented by 33
;; - BC = 0
;;
;; Destroys:
;; - F

Mov33B:	ld	bc, 17
	ldir
Mov16B:	ld	bc, 16
MovNB:	ldir
	ret

;; ShortenBignum:
;;
;; Reduce a TI-format bignum to its canonical form, by discarding
;; leading zeroes.
;;
;; Input:
;; - HL = address of bignum
;;
;; Destroys:
;; - AF, DE

ShortenBignum:
	ld	e, (hl)
	xor	a
	ld	d, a
	cp	e
	ex	de, hl
	add	hl, de
	ex	de, hl
ShortenBignum_Loop:
	ret	z
	ld	a, (de)
	or	a
	ret	nz
	dec	de
	dec	(hl)
	jr	ShortenBignum_Loop

;; HexLToASCII:
;; HexHToASCII:
;;
;; Convert a nibble of A into an ASCII hex digit.
;;
;; Input:
;; - A = value
;;
;; Output:
;; - A = ASCII character
;;
;; Destroys:
;; - F

HexHToASCII:
	rrca
	rrca
	rrca
	rrca
HexLToASCII:
	or	0F0h
	daa
	add	a, 0A0h
	adc	a, 40h
	ret

